@@ -40,11 +40,12 @@
 
 #include <simgear/io/sg_netChat.hxx>
 
-#include <simgear/misc/strutils.hxx>
 #include <simgear/compiler.h>
 #include <simgear/debug/logstream.hxx>
-#include <simgear/timing/timestamp.hxx>
+#include <simgear/misc/strutils.hxx>
 #include <simgear/structure/exception.hxx>
+#include <simgear/timing/timestamp.hxx>
+#include <string>
 
 #include "HTTPClient_private.hxx"
 #include "HTTPTestApi_private.hxx"
@@ -57,6 +58,8 @@
 #  endif
 #endif
 
+using namespace std::string_literals;
+
 namespace simgear
 {
 
@@ -81,21 +84,8 @@ void Client::ClientPrivate::createCurlMulti() {
 #endif
 }
 
-Client::Client() :
-    d(new ClientPrivate)
+Client::Client()
 {
-    d->proxyPort = 0;
-    d->maxConnections = 4;
-    d->maxHostConnections = 4;
-    d->bytesTransferred = 0;
-    d->lastTransferRate = 0;
-    d->timeTransferSample.stamp();
-    d->totalBytesDownloaded = 0;
-    d->maxPipelineDepth = 5;
-    setUserAgent("SimGear-" SG_STRINGIZE(SIMGEAR_VERSION));
-
-    d->tlsCertificatePath = SGPath::fromEnv("SIMGEAR_TLS_CERT_PATH");
-
     static bool didInitCurlGlobal = false;
     static std::mutex initMutex;
 
@@ -105,7 +95,7 @@ Client::Client() :
       didInitCurlGlobal = true;
     }
 
-    d->createCurlMulti();
+    reset();
 }
 
 Client::~Client()
@@ -139,8 +129,22 @@ void Client::setMaxPipelineDepth(unsigned int depth)
 
 void Client::reset()
 {
-    curl_multi_cleanup(d->curlMulti);
+    if (d.get()) {
+        curl_multi_cleanup(d->curlMulti);
+    }
+
     d.reset(new ClientPrivate);
+
+    d->proxyPort = 0;
+    d->maxConnections = 4;
+    d->maxHostConnections = 4;
+    d->bytesTransferred = 0;
+    d->lastTransferRate = 0;
+    d->timeTransferSample.stamp();
+    d->totalBytesDownloaded = 0;
+    d->maxPipelineDepth = 5;
+    d->curlPerformActive = false;
+    setUserAgent("SimGear-" SG_STRINGIZE(SIMGEAR_VERSION));
     d->tlsCertificatePath = SGPath::fromEnv("SIMGEAR_TLS_CERT_PATH");
     d->createCurlMulti();
 }
@@ -162,16 +166,23 @@ void Client::update(int waitTimeout)
         return;
     }
 
+    d->curlPerformActive = true;
     mc = curl_multi_perform(d->curlMulti, &remainingActive);
     if (mc == CURLM_CALL_MULTI_PERFORM) {
         // we could loop here, but don't want to get blocked
         // also this shouldn't  ocurr in any modern libCurl
         curl_multi_perform(d->curlMulti, &remainingActive);
     } else if (mc != CURLM_OK) {
+        d->curlPerformActive = false;
         SG_LOG(SG_IO, SG_WARN, "curl_multi_perform failed:" << curl_multi_strerror(mc));
         return;
     }
 
+    d->curlPerformActive = false;
+    for (auto r : d->pendingCancelRequests) {
+        cancelRequest(r, "deferred cancellation");
+    };
+
     CURLMsg* msg;
     while ((msg = curl_multi_info_read(d->curlMulti, &messagesInQueue))) {
       if (msg->msg == CURLMSG_DONE) {
@@ -202,7 +213,17 @@ void Client::update(int waitTimeout)
           }
 
           if (doProcess) {
-            if (msg->data.result == 0) {
+            bool ok = false;
+            SG_LOG(SG_IO, SG_DEBUG, "msg->data.result=" << msg->data.result);
+            if (msg->data.result == CURLE_OK) {
+              ok = true;
+            }
+            else if (req->range() != "" && (responseCode == 206 || responseCode == 416)) {
+              SG_LOG(SG_IO, SG_DEBUG, "req->range() is set (to '" << req->range() << "')"
+                      << " so treating as ok: msg->data.result=" << msg->data.result);
+              ok = true;
+            }
+            if (ok) {
               req->responseComplete();
             } else {
               SG_LOG(SG_IO, SG_WARN,
@@ -255,6 +276,10 @@ void Client::makeRequest(const Request_ptr& r)
     curl_easy_setopt(curlRequest, CURLOPT_WRITEDATA, r.get());
     curl_easy_setopt(curlRequest, CURLOPT_HEADERFUNCTION, requestHeaderCallback);
     curl_easy_setopt(curlRequest, CURLOPT_HEADERDATA, r.get());
+    unsigned long bps = r->getMaxBytesPerSec();
+    if (bps != 0) {
+        curl_easy_setopt(curlRequest, CURLOPT_MAX_RECV_SPEED_LARGE, bps);
+    }
 
 #if !defined(CURL_MAX_READ_SIZE)
 	const int CURL_MAX_READ_SIZE = 512 * 1024;
@@ -264,7 +289,7 @@ void Client::makeRequest(const Request_ptr& r)
     curl_easy_setopt(curlRequest, CURLOPT_USERAGENT, d->userAgent.c_str());
     curl_easy_setopt(curlRequest, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
 
-    if (sglog().would_log(SG_TERRASYNC, SG_DEBUG)) {
+    if (sglog().would_log(SG_TERRASYNC, SG_DEBUG, __FILE__, __LINE__, __FUNCTION__)) {
         curl_easy_setopt(curlRequest, CURLOPT_VERBOSE, 1);
     }
 
@@ -285,15 +310,27 @@ void Client::makeRequest(const Request_ptr& r)
       }
     }
 
+    std::string range = r->range();
+    if (range != "") {
+        curl_easy_setopt(curlRequest, CURLOPT_RANGE, range.c_str());
+        SG_LOG(SG_GENERAL, SG_DEBUG, "Have added CURLOPT_RANGE: '" << range << "'");
+    }
+    
+    const char* enc = r->getAcceptEncoding();
+    if (enc) {
+        curl_easy_setopt(curlRequest, CURLOPT_ACCEPT_ENCODING, const_cast<char*>(enc));
+        SG_LOG(SG_GENERAL, SG_ALERT, "Have added CURLOPT_ACCEPT_ENCODING: '" << enc << "'");
+    }
+
     const std::string method = strutils::lowercase (r->method());
     if (method == "get") {
-      curl_easy_setopt(curlRequest, CURLOPT_HTTPGET, 1);
+      // see https://curl.se/libcurl/c/CURLOPT_HTTPGET.html
+      curl_easy_setopt(curlRequest, CURLOPT_HTTPGET, 1L);
     } else if (method == "put") {
-      curl_easy_setopt(curlRequest, CURLOPT_PUT, 1);
-      curl_easy_setopt(curlRequest, CURLOPT_UPLOAD, 1);
+      curl_easy_setopt(curlRequest, CURLOPT_UPLOAD, 1L);
     } else if (method == "post") {
       // see http://curl.haxx.se/libcurl/c/CURLOPT_POST.html
-      curl_easy_setopt(curlRequest, CURLOPT_HTTPPOST, 1);
+      curl_easy_setopt(curlRequest, CURLOPT_POST, 1L);
 
       std::string q = r->query().substr(1);
       curl_easy_setopt(curlRequest, CURLOPT_COPYPOSTFIELDS, q.c_str());
@@ -343,9 +380,16 @@ void Client::cancelRequest(const Request_ptr &r, std::string reason)
         return;
     }
 
+    // don't cancel from inside curl_multi_perform, causes
+    // everything to melt. Defer until after it's done
+    if (d->curlPerformActive) {
+        d->pendingCancelRequests.push_back(r);
+        return;
+    }
+
     CURLMcode err = curl_multi_remove_handle(d->curlMulti, it->second);
     if (err != CURLM_OK) {
-      SG_LOG(SG_IO, SG_WARN, "curl_multi_remove_handle failed:" << err);
+        SG_LOG(SG_IO, SG_WARN, "cancelRequest: curl_multi_remove_handle failed:" << err);
     }
 
     // clear the request pointer form the curl-easy object
