@@ -54,12 +54,24 @@ static int boolify(naContext ctx, naRef r)
 {
     if(IS_NUM(r)) return r.num != 0;
     if(IS_NIL(r) || IS_END(r)) return 0;
+
+    // empty vectors are false, non-empty vectors are true
+    if (IS_VEC(r)) {
+        return naVec_size(r) > 0;
+    }
+
+    // empty hashes are false, non-empty are true
+    if (IS_HASH(r)) {
+        return naHash_size(r) > 0;
+    }
+
     if(IS_STR(r)) {
         double d;
         if(naStr_len(r) == 0) return 0;
         if(naStr_tonum(r, &d)) return d != 0;
         else return 1;
     }
+    if (IS_OBJ(r)) return 1;
     ERR(ctx, "non-scalar used in boolean context");
     return 0;
 }
@@ -157,7 +169,7 @@ static void initContext(naContext c)
     c->error[0] = 0;
     c->userData = 0;
 }
-#define BASE_SIZE 256000
+
 static void initGlobals()
 {
     int i;
@@ -168,11 +180,13 @@ static void initGlobals()
     globals->sem = naNewSem();
     globals->lock = naNewLock();
 
-    globals->allocCount = BASE_SIZE; // reasonable starting value
+    globals->allocCount = 256; // reasonable starting value
     for(i=0; i<NUM_NASAL_TYPES; i++)
         naGC_init(&(globals->pools[i]), i);
-    globals->deadsz = BASE_SIZE;
+    globals->deadsz = 256;
     globals->ndead = 0;
+    // REVIEW: Memory Leak - 2,048,000 bytes in 1 blocks are still reachable
+    // naFree(globals->deadBlocks) is required during shutdown
     globals->deadBlocks = naAlloc(sizeof(void*) * globals->deadsz);
 
     // Initialize a single context
@@ -208,6 +222,7 @@ naContext naNewContext()
         initContext(c);
     } else {
         UNLOCK();
+        // REVIEW: Memory Leak - 8,184 bytes in 1 blocks are still reachable
         c = (naContext)naAlloc(sizeof(struct Context));
         initTemps(c);
         initContext(c);
@@ -318,6 +333,24 @@ static void checkNamedArgs(naContext ctx, struct naCode* c, struct naHash* h)
     }
 }
 
+
+/* This is a C equivalent to NasalRefDescription() in
+flightgear/src/Scripting/NasalSys_private.hxx. */
+static char* NasalRefDescription(naRef val)
+{
+    if (naIsNil(val))           return dosprintf("nil");
+    else if (naIsNum(val))      return dosprintf("num: %f", naNumValue(val).num);
+    else if (naIsString(val))   return dosprintf("string: %s", naStr_data(val));
+    else if (naIsScalar(val))   return dosprintf("scalar");
+    else if (naIsVector(val))   return dosprintf("vector");
+    else if (naIsHash(val))     return dosprintf("hash");
+    else if (naIsFunc(val))     return dosprintf("func");
+    else if (naIsCode(val))     return dosprintf("code");
+    else if (naIsCCode(val))    return dosprintf("ccode");
+    else if (naIsGhost(val))    return dosprintf("ghost");
+    else return dosprintf("?");
+}
+
 static struct Frame* setupFuncall(naContext ctx, int nargs, int mcall, int named)
 {
     naRef *args, func, code, obj = naNil();
@@ -326,7 +359,13 @@ static struct Frame* setupFuncall(naContext ctx, int nargs, int mcall, int named
 
     args = &ctx->opStack[opf];
     func = ctx->opStack[--opf];
-    if(!IS_FUNC(func)) ERR(ctx, "function/method call on uncallable object");
+    if(!IS_FUNC(func)) {
+        char* func_description = NasalRefDescription(func);
+        char* description = dosprintf("function/method call on uncallable object: %s", func_description);
+        ERR(ctx, description);
+        naFree(description);
+        naFree(func_description);
+    }
     code = PTR(func).func->code;
     if(mcall) obj = ctx->opStack[--opf];
     ctx->opFrame = opf;
@@ -973,16 +1012,14 @@ naRef naContinue(naContext ctx)
 
 static void logError(naContext ctx)
 {
-    int i, stack_depth = naStackDepth(ctx);
     printf("Nasal runtime error: %s\n", naGetError(ctx));
-    if( stack_depth < 1 )
-        return;
-    printf("  at %s\n", naStr_data(naGetSourceFile(ctx, 0)));
-    printf(", line %d\n", naGetLine(ctx, 0));
-    for(i = 1; i < stack_depth; ++i )
-        printf( "  called from: %s, line %d",
+    int stack_depth = naStackDepth(ctx);
+    for (int i=0; i<stack_depth; ++i) {
+        printf("    %s: %s, line %d\n",
+                i ? "called from" : "at",
                 naStr_data(naGetSourceFile(ctx, i)),
                 naGetLine(ctx, i));
+    }
 }
 
 static naErrorHandler error_handler = &logError;
