@@ -88,7 +88,12 @@ static int findConstantIndex(struct Parser* p, struct Token* t)
         if(t->type == TOK_SYMBOL) c = naInternSymbol(c);
     } else if(t->type == TOK_FUNC) c = newLambda(p, t);
     else if(t->type == TOK_LITERAL) c = naNum(t->num);
-    else naParseError(p, "invalid/non-constant constant", t->line);
+    else {
+        naParseError(p, "invalid/non-constant constant", t->line);
+        /* naParseError() doesn't return, but this stops compiler complaining
+        about <c> not being set. */
+        return -1;
+    }
     return internConstant(p, c);
 }
 
@@ -338,6 +343,20 @@ static void genShortCircuit(struct Parser* p, struct Token* t)
     fixJumpTarget(p, end);
 }
 
+static void genNullishChain(struct Parser* p, struct Token* t)
+{
+    int end;
+    genExpr(p, LEFT(t));
+    emit(p, OP_DUP); // duplicate since OP_EQ will discard
+    emit(p, OP_PUSHNIL);
+    emit(p, OP_EQ); // compare to exact nil, don't boolify()
+    end = emitJump(p, OP_JIFNOTPOP);
+    // if we didn't jump, pop and use right hand side
+    emit(p, OP_POP); // pop the duplicated value
+    genExpr(p, RIGHT(t));
+    fixJumpTarget(p, end);
+}
+
 
 static void genIf(struct Parser* p, struct Token* tif, struct Token* telse)
 {
@@ -375,6 +394,33 @@ static void genQuestion(struct Parser* p, struct Token* t)
     fixJumpTarget(p, jumpEnd);
 }
 
+static void genNullOrMember(struct Parser* p, struct Token* t)
+{
+    int jumpNext, jumpEnd;
+    genExpr(p, LEFT(t));
+    if (!RIGHT(t) || RIGHT(t)->type != TOK_SYMBOL)
+        naParseError(p, "object field not symbol", RIGHT(t)->line);
+
+    emit(p, OP_DUP); // duplicate since OP_EQ will discard
+    emit(p, OP_PUSHNIL);
+    emit(p, OP_EQ); // compare to exact nil, don't boolify()
+
+    // if left is null, jump over the member access
+    jumpNext = emitJump(p, OP_JIFTRUE);
+    emit(p, OP_POP); // pop the comparisom result
+    // object is non-nil here, emit the regular member access
+    emitImmediate(p, OP_MEMBER, findConstantIndex(p, RIGHT(t)));
+    jumpEnd = emitJump(p, OP_JMP);
+    fixJumpTarget(p, jumpNext);
+
+    // pop comparisom result; we don't pop the LEFT(t) expr result here
+    // since we want a NIL as our result, and we know that expression
+    // is qualt to nil.
+    emit(p, OP_POP);
+
+    fixJumpTarget(p, jumpEnd);
+}
+
 static int countList(struct Token* t, int type)
 {
     int n;
@@ -706,11 +752,18 @@ static void genExpr(struct Parser* p, struct Token* t)
         genExpr(p, LEFT(t));
         if(!RIGHT(t) || RIGHT(t)->type != TOK_SYMBOL)
             naParseError(p, "object field not symbol", RIGHT(t)->line);
+
         emitImmediate(p, OP_MEMBER, findConstantIndex(p, RIGHT(t)));
         break;
+    case TOK_NULL_ACCESS:
+        genNullOrMember(p, t);
+        break;
     case TOK_EMPTY: case TOK_NIL:
         emit(p, OP_PUSHNIL);
         break;
+    case TOK_NULL_CHAIN:
+        genNullishChain(p, t);
+        break;
     case TOK_AND: case TOK_OR:
         genShortCircuit(p, t);
         break;
