@@ -38,6 +38,9 @@
 #include <simgear/structure/exception.hxx>
 
 using namespace simgear;
+using std::cerr;
+using std::cout;
+using std::endl;
 
 std::string readFileIntoString(const SGPath& path)
 {
@@ -166,7 +169,8 @@ int parseTest()
 {
     SGPath rootPath = simgear::Dir::current().path();
     rootPath.append("testRoot");
-    pkg::Root* root = new pkg::Root(rootPath, "8.1.12");
+
+    pkg::RootRef root(new pkg::Root(rootPath, "8.1.2"));
     root->setLocale("de");
     pkg::CatalogRef cat = pkg::Catalog::createFromPath(root, SGPath(SRC_DIR "/catalogTest1"));
 
@@ -177,7 +181,9 @@ int parseTest()
     SG_CHECK_EQUAL(cat->description(), "First test catalog");
 
 // check the packages too
-    SG_CHECK_EQUAL(cat->packages().size(), 6);
+    SG_CHECK_EQUAL(cat->packages().size(), 4);
+    SG_CHECK_EQUAL(cat->packages(simgear::pkg::LibraryPackage).size(), 2);
+    SG_CHECK_EQUAL(cat->packages(simgear::pkg::AIModelPackage).size(), 1);
 
     pkg::PackageRef p1 = cat->packages().front();
     SG_CHECK_EQUAL(p1->catalog(), cat.ptr());
@@ -187,6 +193,7 @@ int parseTest()
     SG_CHECK_EQUAL(p1->name(), "Alpha package");
     SG_CHECK_EQUAL(p1->revision(), 8);
     SG_CHECK_EQUAL(p1->fileSizeBytes(), 593);
+    SG_CHECK_EQUAL(p1->type(), simgear::pkg::AircraftPackage);
 
 
     pkg::PackageRef p2 = cat->getPackageById("c172p");
@@ -344,7 +351,6 @@ int parseTest()
     SG_CHECK_EQUAL(urls.size(), 3);
     SG_CHECK_EQUAL(urls.at(1), "http://localhost:2000/mirrorB/b737.tar.gz");
 
-    delete root;
     return EXIT_SUCCESS;
 }
 
@@ -370,7 +376,7 @@ void testAddCatalog(HTTP::Client* cl)
     p.append("org.flightgear.test.catalog1");
     p.append("catalog.xml");
     SG_VERIFY(p.exists());
-    SG_CHECK_EQUAL(root->allPackages().size(), 6);
+    SG_CHECK_EQUAL(root->allPackages().size(), 4);
     SG_CHECK_EQUAL(root->catalogs().size(), 1);
 
     pkg::PackageRef p1 = root->getPackageById("alpha");
@@ -953,7 +959,7 @@ int parseInvalidTest()
 {
     SGPath rootPath = simgear::Dir::current().path();
     rootPath.append("testRoot");
-    pkg::Root* root = new pkg::Root(rootPath, "8.1.12");
+    pkg::RootRef root(new pkg::Root(rootPath, "8.1.12"));
     pkg::CatalogRef cat = pkg::Catalog::createFromPath(root, SGPath(SRC_DIR "/catalogTestInvalid"));
     SG_VERIFY(cat.valid());
 
@@ -1219,8 +1225,8 @@ void testMigrateInstalled(HTTP::Client *cl) {
     waitForUpdateComplete(cl, root);
 
     string_list existing;
-    for (const auto &pack : oldCatalog->installedPackages()) {
-      existing.push_back(pack->id());
+    for (const auto& pack : oldCatalog->installedPackages(simgear::pkg::AnyPackageType)) {
+        existing.push_back(pack->id());
     }
 
     SG_CHECK_EQUAL(4, existing.size());
@@ -1309,6 +1315,31 @@ void testDontMigrateRemoved(HTTP::Client *cl) {
   }
 }
 
+void testProvides(HTTP::Client* cl)
+{
+    SGPath rootPath(simgear::Dir::current().path());
+    rootPath.append("pkg_check_provides");
+    simgear::Dir pd(rootPath);
+    pd.removeChildren();
+    global_catalogVersion = 0;
+
+    pkg::RootRef root(new pkg::Root(rootPath, "8.1.2"));
+    // specify a test dir
+    root->setHTTPClient(cl);
+
+    pkg::CatalogRef c = pkg::Catalog::createFromUrl(root.ptr(), "http://localhost:2000/catalogTest1/catalog.xml");
+    waitForUpdateComplete(cl, root);
+
+    // query
+    auto packages = root->packagesProviding("Aircraft/b744", false);
+    SG_CHECK_EQUAL(packages.size(), 1);
+    SG_CHECK_EQUAL(packages.front()->qualifiedId(), "org.flightgear.test.catalog1.b747-400");
+
+    packages = root->packagesProviding("movies/Foo/intro.mov", false);
+    SG_CHECK_EQUAL(packages.size(), 1);
+    SG_CHECK_EQUAL(packages.front()->qualifiedId(), "org.flightgear.test.catalog1.movies");
+}
+
 int main(int argc, char* argv[])
 {
     sglog().setLogLevels( SG_ALL, SG_WARN );
@@ -1333,31 +1364,33 @@ int main(int argc, char* argv[])
     testRefreshCatalog(&cl);
 
     testInstallTarPackage(&cl);
-    
+
     testInstallArchiveType(&cl);
-    
+
     testDisableDueToVersion(&cl);
-    
+
     testOfflineMode(&cl);
-    
+
     testVersionMigrate(&cl);
-    
+
     updateInvalidToValid(&cl);
     updateValidToInvalid(&cl);
     updateInvalidToInvalid(&cl);
-    
+
     removeInvalidCatalog(&cl);
-    
+
     testVersionMigrateToId(&cl);
 
     testInstallBadPackage(&cl);
-    
+
     testMirrorsFailure(&cl);
 
     testMigrateInstalled(&cl);
 
     testDontMigrateRemoved(&cl);
 
+    testProvides(&cl);
+
     cerr << "Successfully passed all tests!" << endl;
     return EXIT_SUCCESS;
 }
