@@ -54,7 +54,7 @@ public:
         }
 
         selectMirrorUrl();
-        
+
         m_extractPath = aOwner->path().dir();
         m_extractPath.append("_extract_" + aOwner->package()->md5());
 
@@ -96,7 +96,7 @@ protected:
         m_activeURL = m_urls.at(randomizedIndex);
         m_urls.erase(m_urls.begin() + randomizedIndex);
     }
-    
+
     virtual std::string url() const override
     {
         return m_activeURL;
@@ -114,18 +114,18 @@ protected:
 		m_extractor.reset(new ArchiveExtractor(m_extractPath));
         memset(&m_md5, 0, sizeof(SG_MD5_CTX));
         SG_MD5Init(&m_md5);
-        
+
         m_owner->startDownload();
     }
 
     void gotBodyData(const char* s, int n) override
     {
-        // if there's a pre-existing error, discard byte sinstead of pushing
-        // more through the extactor
+        // if there's a pre-existing error, discard bytes instead of pushing
+        // more through the extractor
         if (m_extractor->hasError()) {
             return;
         }
-        
+
 		const uint8_t* ubytes = (uint8_t*) s;
         SG_MD5Update(&m_md5, ubytes, n);
         m_downloaded += n;
@@ -213,8 +213,11 @@ protected:
 
     void onFail() override
     {
+        const auto root = m_owner->package()->catalog()->root();
         if (responseCode() == -1) {
             doFailure(Delegate::USER_CANCELLED);
+        } else if (!root->isOnline()) {
+            doFailure(Delegate::FAIL_OFFLINE);
         } else {
             doFailure(Delegate::FAIL_DOWNLOAD);
         }
@@ -227,14 +230,14 @@ private:
         if (dir.exists()) {
             dir.remove(true /* recursive */);
         }
-        
+
         const auto canRetry = (aReason == Delegate::FAIL_NOT_FOUND) ||
             (aReason == Delegate::FAIL_DOWNLOAD) || (aReason == Delegate::FAIL_CHECKSUM);
-        
+
         if (canRetry && !m_urls.empty()) {
             SG_LOG(SG_GENERAL, SG_WARN, "archive download failed from:" << m_activeURL
                    << "\n\twill retry with next mirror");
-            // becuase selectMirrorUrl erased the active URL from m_urls,
+            // because selectMirrorUrl erased the active URL from m_urls,
             // this new request will select one of the other mirrors
             auto retryDownload = new PackageArchiveDownloader(m_owner, m_urls);
             m_owner->m_download.reset(retryDownload);
@@ -246,9 +249,9 @@ private:
         m_owner->installResult(aReason);
     }
 
-    
+
     InstallRef m_owner;
-    string m_activeURL;
+    std::string m_activeURL;
     string_list m_urls;
     SG_MD5_CTX m_md5;
     SGPath m_extractPath;
