@@ -1,24 +1,18 @@
-// condition.cxx - Declarations and inline methods for property conditions.
-//
-// Written by David Megginson, started 2000.
-// CLO May 2003 - Split out condition specific code.
-//
-// This file is in the Public Domain, and comes with no warranty.
-//
-// $Id$
+// SPDX-FileName: condition.cxx
+// SPDX-License-Identifier: LGPL-2.1-only
+// SPDX-FileComment: Declarations and inline methods for property conditions
 
-#ifdef HAVE_CONFIG_H
-#  include <simgear_config.h>
-#endif
-
-// #include <iostream>
+#include <simgear_config.h>
 
+#include <simgear/debug/ErrorReportingCallback.hxx>
+#include <simgear/debug/logstream.hxx>
+#include <simgear/structure/SGExpression.hxx>
 #include <simgear/structure/exception.hxx>
 
 #include "props.hxx"
 #include "condition.hxx"
+#include <simgear/misc/strutils.hxx>
 
-#include <simgear/structure/SGExpression.hxx>
 
 using std::istream;
 using std::ostream;
@@ -235,6 +229,7 @@ SGAndCondition::test () const
 void
 SGAndCondition::addCondition (SGCondition * condition)
 {
+  // REVIEW: Memory Leak - 632 bytes in 79 blocks are indirectly lost
   _conditions.push_back(condition);
 }
 
@@ -311,9 +306,8 @@ doComparison (const SGPropertyNode * left, const SGPropertyNode * right, const S
       return SGComparisonCondition::LESS_THAN;
     else if (v1 > v2)
       return SGComparisonCondition::GREATER_THAN;
-    else
-      return SGComparisonCondition::EQUALS;
-    break;
+
+    return SGComparisonCondition::EQUALS;
   }
   case props::INT:
     return doComp<int>(left->getIntValue(), right->getIntValue(), 
@@ -341,9 +335,8 @@ doComparison (const SGPropertyNode * left, const SGPropertyNode * right, const S
       return SGComparisonCondition::LESS_THAN;
     else if (v1 > v2)
       return SGComparisonCondition::GREATER_THAN;
-    else
-      return SGComparisonCondition::EQUALS;
-    break;
+
+    return SGComparisonCondition::EQUALS;
   }
   default:
     throw sg_exception("condition: unrecognized node type in comparison");
@@ -426,6 +419,7 @@ SGComparisonCondition::setPrecisionValue (const SGPropertyNode *node)
 void
 SGComparisonCondition::setRightValue (const SGPropertyNode *node)
 {
+  // REVIEW: Memory Leak - 7,144 bytes in 47 blocks are indirectly lost
   _right_property = new SGPropertyNode(*node);
 }
 
@@ -482,7 +476,8 @@ static SGCondition *
 readPropertyCondition( SGPropertyNode *prop_root,
                        const SGPropertyNode *node )
 {
-  return new SGPropertyCondition( prop_root, node->getStringValue() );
+  const auto s = simgear::strutils::strip(node->getStringValue());
+  return new SGPropertyCondition( prop_root, s.c_str());
 }
 
 static SGCondition *
@@ -502,6 +497,7 @@ readNotCondition( SGPropertyNode *prop_root, const SGPropertyNode *node )
 static SGCondition *
 readAndConditions( SGPropertyNode *prop_root, const SGPropertyNode *node )
 {
+  // REVIEW: Memory Leak - 11,502 (1,560 direct, 9,942 indirect) bytes in 39 blocks are definitely lost
   SGAndCondition * andCondition = new SGAndCondition;
   int nChildren = node->nChildren();
   for (int i = 0; i < nChildren; i++) {
@@ -533,54 +529,59 @@ readComparison( SGPropertyNode *prop_root,
                 SGComparisonCondition::Type type,
 		bool reverse)
 {
+  // REVIEW: Memory Leak - 2,880 bytes in 40 blocks are indirectly lost
   SGComparisonCondition * condition = new SGComparisonCondition(type, reverse);
   if (node->nChildren() < 2 || node->nChildren() > 3 ) {
-    throw sg_exception("condition: comparison without two or three children");
+    throw sg_exception("condition: comparison without two or three children",
+                       {}, {}, false);
   }
   
   const SGPropertyNode* left = node->getChild(0), 
     *right = node->getChild(1);
 
   {
-    string leftName(left->getName());
+    string leftName(left->getNameString());
     if (leftName == "property") {
-      condition->setLeftProperty(prop_root, left->getStringValue());
+      condition->setLeftProperty(prop_root, left->getStringValue().c_str());
     } else if (leftName == "value") {
       condition->setLeftValue(left);
     } else if (leftName == "expression") {
       SGExpressiond* exp = SGReadDoubleExpression(prop_root, left->getChild(0));
       condition->setLeftDExpression(exp);
     } else {
-      throw sg_exception("Unknown condition comparison left child:" + leftName);
+      throw sg_exception("Unknown condition comparison left child:" + leftName,
+                         {}, {}, false);
     }
   }
     
   {
-    string rightName(right->getName());
+    string rightName(right->getNameString());
     if (rightName == "property") {
-      condition->setRightProperty(prop_root, right->getStringValue());
+      condition->setRightProperty(prop_root, right->getStringValue().c_str());
     } else if (rightName == "value") {
       condition->setRightValue(right);
     } else if (rightName == "expression") {
       SGExpressiond* exp = SGReadDoubleExpression(prop_root, right->getChild(0));
       condition->setRightDExpression(exp);
     } else {
-      throw sg_exception("Unknown condition comparison right child:" + rightName);
+      throw sg_exception("Unknown condition comparison right child:" + rightName,
+                         {}, {}, false);
     }
   }
   
   if( node->nChildren() == 3 ) {
     const SGPropertyNode *n = node->getChild(2);
-    string name(n->getName());
+    string name(n->getNameString());
     if (name == "precision-property") {
-      condition->setPrecisionProperty(prop_root, n->getStringValue());
+      condition->setPrecisionProperty(prop_root, n->getStringValue().c_str());
     } else if (name == "precision-value") {
       condition->setPrecisionValue(n);
     } else if (name == "precision-expression") {
       SGExpressiond* exp = SGReadDoubleExpression(prop_root, n->getChild(0));
       condition->setPrecisionDExpression(exp);
     } else {
-      throw sg_exception("Unknown condition comparison precision child:" + name );
+      throw sg_exception("Unknown condition comparison precision child:" + name,
+                         {}, {}, false);
     }
   }
   
@@ -590,7 +591,7 @@ readComparison( SGPropertyNode *prop_root,
 static SGCondition *
 readCondition( SGPropertyNode *prop_root, const SGPropertyNode *node )
 {
-  const string &name = node->getName();
+  const string &name = node->getNameString();
   if (name == "property")
     return readPropertyCondition(prop_root, node);
   else if (name == "not")
@@ -655,6 +656,28 @@ SGConditional::test () const
 SGCondition *
 sgReadCondition( SGPropertyNode *prop_root, const SGPropertyNode *node )
 {
+  // detect the case where someone creates a condition with no property children
+  // this typically means the user wrote either a literal value or a property path.
+  // unfortunately we can't decide what to do here.
+  if (node->nChildren() == 0) {
+    // always report the error
+    simgear::reportFailure(simgear::LoadFailure::Misconfigured,
+                           simgear::ErrorCode::XMLModelLoad,
+                           "sgReadCondition: condition has no children, textual value is:" + node->getStringValue() +
+                               ", condition will always be false.",
+                           sg_location{node});
+
+    // aircraft developer mode: throw an exception, will become the default behaviour
+    // in the future
+    if (sglog().inDeveloperMode()) {
+      throw sg_exception("condition with no children: textual value is:" + node->getStringValue(),
+                         "sgReadCondition", sg_location{node}, false /* don't report, we did it above */);
+    } else {
+      // legacy compat behaviour: create an empty and-expression which therefore evaluates to 'true'
+      return new SGAndCondition;
+    }
+  }
+
   return readAndConditions(prop_root, node);
 }
 
