@@ -30,6 +30,7 @@
 #include <osg/TexGen>
 #include <osg/Texture1D>
 #include <osg/Texture2D>
+#include <osg/Texture2DArray>
 #include <osg/Texture3D>
 #include <osg/TextureRectangle>
 #include <osg/TextureCubeMap>
@@ -45,7 +46,7 @@
 #include <simgear/scene/util/SGReaderWriterOptions.hxx>
 #include <simgear/scene/util/SGSceneFeatures.hxx>
 #include <simgear/scene/util/StateAttributeFactory.hxx>
-#include <simgear/structure/OSGUtils.hxx>
+#include <simgear/scene/util/OsgUtils.hxx>
 
 namespace simgear
 {
@@ -64,12 +65,12 @@ TexEnvCombine* buildTexEnvCombine(Effect* effect,
 TexGen* buildTexGen(Effect* Effect, const SGPropertyNode* tgenProp);
 
 // Hack to force inclusion of TextureBuilder.cxx in library
-osg::Texture* TextureBuilder::buildFromType(Effect* effect, Pass* pass, const string& type,
+osg::Texture* TextureBuilder::buildFromType(Effect* effect, const string& type,
                                             const SGPropertyNode*props,
                                             const SGReaderWriterOptions*
                                             options)
 {
-    return EffectBuilder<Texture>::buildFromType(effect, pass, type, props, options);
+    return EffectBuilder<Texture>::buildFromType(effect, type, props, options);
 }
 
 typedef std::tuple<string, Texture::FilterMode, Texture::FilterMode,
@@ -136,7 +137,7 @@ void TextureUnitBuilder::buildAttribute(Effect* effect, Pass* pass,
         type = pType->getStringValue();
     Texture* texture = 0;
     try {
-        texture = TextureBuilder::buildFromType(effect, pass, type, prop,
+        texture = TextureBuilder::buildFromType(effect, type, prop,
                                                 options);
     }
     catch (BuilderException& e) {
@@ -276,11 +277,7 @@ bool setAttrs(const TexTuple& attrs, Texture* tex,
         options->setLoadOriginHint(SGReaderWriterOptions::LoadOriginHint::ORIGIN_EFFECTS);
 
     try {
-    #if OSG_VERSION_LESS_THAN(3,4,2)
-        result = osgDB::readImageFile(imageName, options);
-    #else
         result = osgDB::readRefImageFile(imageName, options);
-    #endif
     } catch (std::exception& e) {
         simgear::reportFailure(simgear::LoadFailure::OutOfMemory, simgear::ErrorCode::LoadingTexture,
                                string{"osgDB::readRefImageFile failed:"} + e.what(),
@@ -328,7 +325,7 @@ class TexBuilder : public TextureBuilder
 {
 public:
     TexBuilder(const string& texType) : _type(texType) {}
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 protected:
     typedef map<TexTuple, observer_ptr<T> > TexMap;
@@ -337,7 +334,7 @@ protected:
 };
 
 template<typename T>
-Texture* TexBuilder<T>::build(Effect* effect, Pass* pass, const SGPropertyNode* props,
+Texture* TexBuilder<T>::build(Effect* effect, const SGPropertyNode* props,
                               const SGReaderWriterOptions* options)
 {
     TexTuple attrs = makeTexTuple(effect, props, options, _type);
@@ -372,11 +369,11 @@ TextureBuilder::Registrar install2D("2d", new TexBuilder<Texture2D>("2d"));
 class WhiteTextureBuilder : public TextureBuilder
 {
 public:
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 };
 
-Texture* WhiteTextureBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode*,
+Texture* WhiteTextureBuilder::build(Effect* effect, const SGPropertyNode*,
                                     const SGReaderWriterOptions* options)
 {
     return StateAttributeFactory::instance()->getWhiteTexture();
@@ -390,11 +387,11 @@ TextureBuilder::Registrar installWhite("white", new WhiteTextureBuilder);
 class TransparentTextureBuilder : public TextureBuilder
 {
 public:
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 };
 
-Texture* TransparentTextureBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode*,
+Texture* TransparentTextureBuilder::build(Effect* effect, const SGPropertyNode*,
                                     const SGReaderWriterOptions* options)
 {
     return StateAttributeFactory::instance()->getTransparentTexture();
@@ -406,17 +403,36 @@ TextureBuilder::Registrar installTransparent("transparent",
                                              new TransparentTextureBuilder);
 }
 
+class NullNormalmapTextureBuilder : public TextureBuilder
+{
+public:
+    Texture* build(Effect* effect, const SGPropertyNode*,
+                   const SGReaderWriterOptions* options);
+};
+
+Texture* NullNormalmapTextureBuilder::build(Effect* effect, const SGPropertyNode*,
+                                            const SGReaderWriterOptions* options)
+{
+    return StateAttributeFactory::instance()->getNullNormalmapTexture();
+}
+
+namespace
+{
+TextureBuilder::Registrar installNullNormalmap("null-normalmap",
+                                               new NullNormalmapTextureBuilder);
+}
+
 class NoiseBuilder : public TextureBuilder
 {
 public:
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 protected:
     typedef map<int, ref_ptr<Texture3D> > NoiseMap;
     NoiseMap _noises;
 };
 
-Texture* NoiseBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode* props,
+Texture* NoiseBuilder::build(Effect* effect, const SGPropertyNode* props,
                              const SGReaderWriterOptions* options)
 {
     int texSize = 64;
@@ -436,7 +452,7 @@ TextureBuilder::Registrar installNoise("noise", new NoiseBuilder);
 class LightSpriteBuilder : public TextureBuilder
 {
 public:
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 protected:
     Mutex lightMutex;
@@ -445,7 +461,7 @@ protected:
     osg::Image* getPointSpriteImage(int logResolution);
 };
 
-Texture* LightSpriteBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode* props,
+Texture* LightSpriteBuilder::build(Effect* effect, const SGPropertyNode* props,
                              const SGReaderWriterOptions* options)
 {
     ScopedLock<Mutex> lock(lightMutex);
@@ -563,7 +579,7 @@ CubeMapTuple makeCubeMapTuple(Effect* effect, const SGPropertyNode* props)
 class CubeMapBuilder : public TextureBuilder
 {
 public:
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 protected:
     typedef map<CubeMapTuple, observer_ptr<TextureCubeMap> > CubeMap;
@@ -586,7 +602,7 @@ void copySubImage(const osg::Image* srcImage, int src_s, int src_t, int width, i
 }
 
 
-Texture* CubeMapBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode* props,
+Texture* CubeMapBuilder::build(Effect* effect, const SGPropertyNode* props,
                                const SGReaderWriterOptions* options)
 {
     // First check that there is a <images> tag
@@ -620,56 +636,34 @@ Texture* CubeMapBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode*
         SGReaderWriterOptions* wOpts = (SGReaderWriterOptions*)options;
         SGReaderWriterOptions::LoadOriginHint origLOH = wOpts->getLoadOriginHint();
         wOpts->setLoadOriginHint(SGReaderWriterOptions::LoadOriginHint::ORIGIN_EFFECTS);
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(std::get<0>(_tuple), options);
-#else
         result = osgDB::readRefImageFile(std::get<0>(_tuple), options);
-#endif
+
         if(result.success()) {
             osg::Image* image = result.getImage();
             cubeTexture->setImage(TextureCubeMap::POSITIVE_X, image);
         }
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(std::get<1>(_tuple), options);
-#else
         result = osgDB::readRefImageFile(std::get<1>(_tuple), options);
-#endif
         if(result.success()) {
             osg::Image* image = result.getImage();
             cubeTexture->setImage(TextureCubeMap::NEGATIVE_X, image);
         }
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(std::get<2>(_tuple), options);
-#else
         result = osgDB::readRefImageFile(std::get<2>(_tuple), options);
-#endif
         if(result.success()) {
             osg::Image* image = result.getImage();
             cubeTexture->setImage(TextureCubeMap::POSITIVE_Y, image);
         }
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(std::get<3>(_tuple), options);
-#else
         result = osgDB::readRefImageFile(std::get<3>(_tuple), options);
-#endif
+
         if(result.success()) {
             osg::Image* image = result.getImage();
             cubeTexture->setImage(TextureCubeMap::NEGATIVE_Y, image);
         }
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(std::get<4>(_tuple), options);
-#else
         result = osgDB::readRefImageFile(std::get<4>(_tuple), options);
-#endif
         if(result.success()) {
             osg::Image* image = result.getImage();
             cubeTexture->setImage(TextureCubeMap::POSITIVE_Z, image);
         }
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(std::get<5>(_tuple), options);
-#else
         result = osgDB::readRefImageFile(std::get<5>(_tuple), options);
-#endif
         if(result.success()) {
             osg::Image* image = result.getImage();
             cubeTexture->setImage(TextureCubeMap::NEGATIVE_Z, image);
@@ -694,11 +688,8 @@ Texture* CubeMapBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode*
             return cubeTexture.release();
 
         osgDB::ReaderWriter::ReadResult result;
-#if OSG_VERSION_LESS_THAN(3,4,1)
-        result = osgDB::readImageFile(texname, options);
-#else
         result = osgDB::readRefImageFile(texname, options);
-#endif
+
         if(result.success()) {
             osg::Image* image = result.getImage();
             image->flipVertical();   // Seems like the image coordinates are somewhat funny, flip to get better ones
@@ -790,17 +781,94 @@ TextureBuilder::Registrar installCubeMap("cubemap", new CubeMapBuilder);
 }
 
 
+typedef map<int, string> Tex2DArraySignature;
+
+Tex2DArraySignature makeTex2DArraySignature(Effect *effect,
+                                            const SGPropertyNode *props)
+{
+    Tex2DArraySignature signature;
+    const PropertyList images = props->getChildren("image");
+    for (const auto &image : images) {
+        const SGPropertyNode *real_prop = getEffectPropertyNode(effect, image);
+        signature[image->getIndex()] = real_prop->getStringValue();
+    }
+    return signature;
+}
+
+class Texture2DArrayBuilder : public TextureBuilder
+{
+public:
+    Texture* build(Effect* effect, const SGPropertyNode*,
+                   const SGReaderWriterOptions* options);
+protected:
+    typedef map<Tex2DArraySignature, observer_ptr<Texture2DArray>> TexMap;
+    TexMap texMap;
+};
+
+Texture* Texture2DArrayBuilder::build(Effect* effect,
+                                      const SGPropertyNode* props,
+                                      const SGReaderWriterOptions* options)
+{
+    Tex2DArraySignature signature = makeTex2DArraySignature(effect, props);
+
+    auto itr = texMap.find(signature);
+    ref_ptr<Texture2DArray> tex;
+
+    if (itr != texMap.end() && itr->second.lock(tex))
+        return tex.release();
+
+    tex = new osg::Texture2DArray;
+
+    osgDB::ReaderWriter::ReadResult result;
+    SGReaderWriterOptions* wOpts = (SGReaderWriterOptions*)options;
+    SGReaderWriterOptions::LoadOriginHint origLOH = wOpts->getLoadOriginHint();
+    wOpts->setLoadOriginHint(SGReaderWriterOptions::LoadOriginHint::ORIGIN_EFFECTS);
+
+    for (const auto &pair : signature) {
+        osg::ref_ptr<osg::Image> image;
+        result = osgDB::readRefImageFile(pair.second, options);
+        if (result.success())
+            image = result.getImage();
+
+        if (image.valid()) {
+            tex->setImage(pair.first, image);
+        } else {
+            SG_LOG(SG_INPUT, SG_ALERT, "failed to load effect texture file '"
+                   << pair.second << "'");
+            simgear::reportFailure(simgear::LoadFailure::BadData, simgear::ErrorCode::LoadingTexture,
+                                   "osgDB::readRefImageFile failed:" + result.message(),
+                                   SGPath::fromUtf8(pair.second));
+            return nullptr;
+        }
+    }
+
+    wOpts->setLoadOriginHint(origLOH);
+
+    // TODO: Filtering, wrap-mode, etc. should be configurable
+
+    if (itr == texMap.end())
+        texMap[signature] = tex;
+    else
+        itr->second = tex; // update existing
+    return tex.release();
+}
+
+namespace {
+TextureBuilder::Registrar install2DArray("2d-array", new Texture2DArrayBuilder);
+}
+
+
 class Texture3DBuilder : public TextureBuilder
 {
 public:
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
+    Texture* build(Effect* effect, const SGPropertyNode*,
                    const SGReaderWriterOptions* options);
 protected:
     typedef map<TexTuple, observer_ptr<Texture3D> > TexMap;
     TexMap texMap;
 };
 
-Texture* Texture3DBuilder::build(Effect* effect, Pass* pass,
+Texture* Texture3DBuilder::build(Effect* effect,
                                  const SGPropertyNode* props,
                                  const SGReaderWriterOptions* options)
 {
@@ -828,11 +896,7 @@ Texture* Texture3DBuilder::build(Effect* effect, Pass* pass,
         options->setLoadOriginHint(SGReaderWriterOptions::LoadOriginHint::ORIGIN_EFFECTS_NORMALIZED);
     else
         options->setLoadOriginHint(SGReaderWriterOptions::LoadOriginHint::ORIGIN_EFFECTS);
-#if OSG_VERSION_LESS_THAN(3,4,2)
-    result = osgDB::readImageFile(imageName, options);
-#else
     result = osgDB::readRefImageFile(imageName, options);
-#endif
     options->setLoadOriginHint(origLOH);
     osg::ref_ptr<osg::Image> image;
     if (result.success())
@@ -1060,7 +1124,7 @@ TexGen* buildTexGen(Effect* effect, const SGPropertyNode* tgenProp)
         for (int i = 0; i < planesNode->nChildren(); ++i) {
             const SGPropertyNode* planeNode = planesNode->getChild(i);
             TexGen::Coord coord;
-            findAttr(tgenCoords, planeNode->getName(), coord);
+            findAttr(tgenCoords, planeNode->getNameString(), coord);
             const SGPropertyNode* realNode
                 = getEffectPropertyNode(effect, planeNode);
             SGVec4d plane = realNode->getValue<SGVec4d>();
@@ -1108,79 +1172,4 @@ bool makeTextureParameters(SGPropertyNode* paramRoot, const StateSet* ss)
     return true;
 }
 
-class GBufferBuilder : public TextureBuilder
-{
-public:
-    GBufferBuilder() {}
-    Texture* build(Effect* effect, Pass* pass, const SGPropertyNode*,
-                   const SGReaderWriterOptions* options);
-private:
-    string buffer;
-};
-
-class BufferNameChangeListener : public SGPropertyChangeListener,
-      public DeferredPropertyListener {
-public:
-    BufferNameChangeListener(Pass* p, int u, const std::string& pn) : pass(p), unit(u)
-    {
-        propName = new std::string(pn);
-    }
-    ~BufferNameChangeListener()
-    {
-        delete propName;
-        propName = 0;
-    }
-    void valueChanged(SGPropertyNode* node)
-    {
-        const char* buffer = node->getStringValue();
-        pass->setBufferUnit(unit, buffer);
-    }
-    void activate(SGPropertyNode* propRoot)
-    {
-        SGPropertyNode* listenProp = makeNode(propRoot, *propName);
-        delete propName;
-        propName = 0;
-        if (listenProp)
-            listenProp->addChangeListener(this, true);
-    }
-
-private:
-    ref_ptr<Pass> pass;
-    int unit;
-    std::string* propName;
-};
-
-Texture* GBufferBuilder::build(Effect* effect, Pass* pass, const SGPropertyNode* prop,
-                                    const SGReaderWriterOptions* options)
-{
-    int unit = 0;
-    const SGPropertyNode* pUnit = prop->getChild("unit");
-    if (pUnit) {
-        unit = pUnit->getValue<int>();
-    } else {
-        SG_LOG(SG_INPUT, SG_ALERT, "no texture unit");
-    }
-    const SGPropertyNode* nameProp = getEffectPropertyChild(effect, prop,
-                                                            "name");
-    if (!nameProp)
-        return 0;
-
-    if (nameProp->nChildren() == 0) {
-        buffer = nameProp->getStringValue();
-        pass->setBufferUnit( unit, buffer );
-    } else {
-        std::string propName = getGlobalProperty(nameProp, options);
-        BufferNameChangeListener* listener = new BufferNameChangeListener(pass, unit, propName);
-        effect->addDeferredPropertyListener(listener);
-    }
-
-    // Return white for now. Would be overridden in Technique::ProcessDrawable
-    return StateAttributeFactory::instance()->getWhiteTexture();
-}
-
-namespace
-{
-    TextureBuilder::Registrar installBuffer("buffer", new GBufferBuilder);
-}
-
 }
