@@ -96,8 +96,8 @@ void SGMaterial::_internal_state::add_texture(const std::string &t, int i)
 SGMaterial::SGMaterial( const SGReaderWriterOptions* options,
                         const SGPropertyNode *props,
                         SGPropertyNode *prop_root,
-                        AreaList *a,
-                        SGSharedPtr<const SGCondition> c,
+                        std::shared_ptr<AreaList> a,
+			SGSharedPtr<const SGCondition> c,
                         const std::string& n)
 {
     init();
@@ -111,7 +111,7 @@ SGMaterial::SGMaterial( const SGReaderWriterOptions* options,
 SGMaterial::SGMaterial( const osgDB::Options* options,
                         const SGPropertyNode *props,
                         SGPropertyNode *prop_root,
-                        AreaList *a,
+                        std::shared_ptr<AreaList> a,
                         SGSharedPtr<const SGCondition> c,
                         const std::string& n)
 {
@@ -129,7 +129,7 @@ SGMaterial::~SGMaterial (void)
 {
 }
 
-
+
 ////////////////////////////////////////////////////////////////////////
 // Public methods.
 ////////////////////////////////////////////////////////////////////////
@@ -230,11 +230,7 @@ SGMaterial::read_properties(const SGReaderWriterOptions* options,
             }
             else
             {
-#if OSG_VERSION_LESS_THAN(3,4,0)
-                osg::Image* image = osgDB::readImageFile(fullMaskPath, options);
-#else
                 osg::Image* image = osgDB::readRefImageFile(fullMaskPath, options);
-#endif
                 if (image && image->valid())
                 {
                     Texture2DRef object_mask = new osg::Texture2D;
@@ -275,25 +271,56 @@ SGMaterial::read_properties(const SGReaderWriterOptions* options,
     mipmap = props->getBoolValue("mipmap", true);
     light_coverage = props->getDoubleValue("light-coverage", 0.0);
 
+    light_edge_spacing_m = props->getDoubleValue("light-edge-spacing-m", 0.0);
+    light_edge_size_cm = props->getDoubleValue("light-edge-size-cm", 40.0);
+    light_edge_height_m = props->getDoubleValue("light-edge-height-m", 5.0);
+    light_edge_intensity_cd = props->getDoubleValue("light-edge-intensity-cd", 50.0);
+    light_edge_angle_horizontal_deg = props->getDoubleValue("light-edge-angle-horizontal-deg", 360.0);
+    light_edge_angle_vertical_deg = props->getDoubleValue("light-edge-angle-vertical-deg", 360.0);
+    light_edge_offset = props->getBoolValue("light-edge-offset", true);
+
+    light_edge_colour[0] = props->getDoubleValue("light-edge-color/r", 1.0);
+    light_edge_colour[1] = props->getDoubleValue("light-edge-color/g", 1.0);
+    light_edge_colour[2] = props->getDoubleValue("light-edge-color/b", 1.0);
+    light_edge_colour[3] = props->getDoubleValue("light-edge-color/a", 1.0);
+
+    // Line feature texture coordinates
+    line_feature_tex_x0 = props->getDoubleValue("line-feature-tex-x0", 0.0);
+    line_feature_tex_x1 = props->getDoubleValue("line-feature-tex-x1", 0.0);
+    line_feature_offset_m = props->getDoubleValue("line-feature-offset-m", 1.0);
+
     // Building properties
     building_coverage = props->getDoubleValue("building-coverage", 0.0);
     building_spacing = props->getDoubleValue("building-spacing-m", 5.0);
 
-    std::string bt = props->getStringValue( "building-texture",
-                                            "Textures/buildings.png" );
+    std::string bt = props->getStringValue("building-texture", "Textures/Buildings/buildings.png");
     building_texture = SGModelLib::findDataFile(bt, options);
 
     if (building_texture.empty()) {
         SG_LOG(SG_GENERAL, SG_ALERT, "Cannot find texture \"" << bt);
     }
 
-    bt = props->getStringValue("building-lightmap", "Textures/buildings-lightmap.png");
+    bt = props->getStringValue("building-lightmap", "Textures/Buildings/buildings-lightmap.png");
     building_lightmap = SGModelLib::findDataFile(bt, options);
 
     if (building_lightmap.empty()) {
         SG_LOG(SG_GENERAL, SG_ALERT, "Cannot find texture \"" << bt);
     }
 
+    bt = props->getStringValue("building-normalmap", "Textures/Buildings/buildings-normalmap.png");
+    building_normalmap = SGModelLib::findDataFile(bt, options);
+
+    if (building_normalmap.empty()) {
+        SG_LOG(SG_GENERAL, SG_ALERT, "Cannot find texture \"" << bt);
+    }
+
+    bt = props->getStringValue("building-reflectionmap", "Textures/Buildings/buildings-reflectionmap.png");
+    building_reflectionmap = SGModelLib::findDataFile(bt, options);
+
+    if (building_reflectionmap.empty()) {
+        SG_LOG(SG_GENERAL, SG_ALERT, "Cannot find texture \"" << bt);
+    }
+
     building_small_ratio = props->getDoubleValue("building-small-ratio", 0.8);
     building_medium_ratio = props->getDoubleValue("building-medium-ratio", 0.15);
     building_large_ratio =  props->getDoubleValue("building-large-ratio", 0.05);
@@ -371,7 +398,13 @@ SGMaterial::read_properties(const SGReaderWriterOptions* options,
     }
 
     // surface values for use with ground reactions
-    _solid = props->getBoolValue("solid", _solid);
+    auto propval = props->getStringValue("solid/property", "");
+    if (propval != "") {
+        _solid_property = prop_root->getNode(propval, true);
+        _solid_is_prop  =true;
+    } else {
+        _solid = props->getBoolValue("solid", _solid);
+    }
     _friction_factor = props->getDoubleValue("friction-factor", _friction_factor);
     _rolling_friction = props->getDoubleValue("rolling-friction", _rolling_friction);
     _bumpiness = props->getDoubleValue("bumpiness", _bumpiness);
@@ -411,9 +444,8 @@ SGMaterial::read_properties(const SGReaderWriterOptions* options,
     // read glyph table for taxi-/runway-signs
     std::vector<SGPropertyNode_ptr> glyph_nodes = props->getChildren("glyph");
     for (unsigned int i = 0; i < glyph_nodes.size(); i++) {
-        const char *name = glyph_nodes[i]->getStringValue("name");
-        if (name)
-            glyphs[name] = new SGMaterialGlyph(glyph_nodes[i]);
+        std::string name = glyph_nodes[i]->getStringValue("name");
+        glyphs[name] = new SGMaterialGlyph(glyph_nodes[i]);
     }
 
     // Read parameters entry, which is passed into the effect
@@ -425,7 +457,6 @@ SGMaterial::read_properties(const SGReaderWriterOptions* options,
 }
 
 
-
 ////////////////////////////////////////////////////////////////////////
 // Private methods.
 ////////////////////////////////////////////////////////////////////////
@@ -441,6 +472,16 @@ SGMaterial::init ()
 
     mipmap = true;
     light_coverage = 0.0;
+
+    light_edge_spacing_m = 0.0;
+    light_edge_size_cm = 50.0;
+    light_edge_height_m = 5.0;
+    light_edge_intensity_cd = 100.0;
+    light_edge_angle_horizontal_deg = 360.0;
+    light_edge_angle_vertical_deg = 360.0;
+    light_edge_colour = SGVec4f(1.0,1.0,1.0,1.0); 
+    light_edge_offset = true;
+
     building_coverage = 0.0;
 
     shininess = 1.0;
@@ -500,6 +541,39 @@ osg::Texture2D* SGMaterial::get_one_object_mask(int texIndex)
     }
 }
 
+std::string SGMaterial::get_one_texture(int setIndex, int texIndex)
+{
+    if (_status.empty()) {
+        SG_LOG( SG_GENERAL, SG_WARN, "No material available.");
+        return 0;
+    }
+
+    unsigned int i = setIndex % _status.size();
+
+    SGMaterial::_internal_state st = _status[i];
+    //if (st == NULL) return 0;
+
+    std::string texturePath = "";
+
+    auto iter = st.texture_paths.begin();
+    for (; iter != st.texture_paths.end(); ++iter) {
+        if (iter->second == texIndex) texturePath = iter->first;
+    }
+
+    return texturePath;
+}
+
+std::size_t SGMaterial::get_num_textures(int setIndex)
+{
+    if (_status.empty()) {
+        SG_LOG( SG_GENERAL, SG_WARN, "No material available.");
+        return 0;
+    }
+
+    SGMaterial::_internal_state st = _status[setIndex % _status.size()];
+    return st.texture_paths.size();
+}
+
 void SGMaterial::buildEffectProperties(const SGReaderWriterOptions* options)
 {
     using namespace osg;
@@ -597,7 +671,6 @@ bool SGMaterial::valid(SGVec2f loc) const
 	return false;
 }
 
-
 ////////////////////////////////////////////////////////////////////////
 // SGMaterialGlyph.
 ////////////////////////////////////////////////////////////////////////
