@@ -17,15 +17,14 @@
 // along with this program; if not, write to the Free Software
 // Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
-#ifdef HAVE_CONFIG_H
-#  include <simgear_config.h>
-#endif
+#include <simgear_config.h>
 
 #include "ModelRegistry.hxx"
 #include <simgear/scene/util/SGImageUtils.hxx>
 #include "../material/mipmap.hxx"
 
 #include <algorithm>
+#include <chrono>
 #include <utility>
 #include <vector>
 
@@ -48,6 +47,8 @@
 #include <osgUtil/Optimizer>
 #include <osg/Texture>
 
+#include <osgTerrain/Terrain>
+
 #include <simgear/sg_inlines.h>
 
 #include <simgear/scene/util/SGSceneFeatures.hxx>
@@ -55,12 +56,14 @@
 #include <simgear/scene/util/SGTextureStateAttributeVisitor.hxx>
 #include <simgear/scene/util/SGReaderWriterOptions.hxx>
 #include <simgear/scene/util/NodeAndDrawableVisitor.hxx>
+#include <simgear/debug/Reporting.hxx>
 
+#include <simgear/scene/tgdb/VPBTechnique.hxx>
 
 #include <simgear/debug/ErrorReportingCallback.hxx>
-
 #include <simgear/io/sg_file.hxx>
 #include <simgear/misc/lru_cache.hxx>
+#include <simgear/misc/strutils.hxx>
 #include <simgear/props/condition.hxx>
 #include <simgear/props/props.hxx>
 #include <simgear/props/props_io.hxx>
@@ -302,6 +305,7 @@ ModelRegistry::readImage(const string& fileName,
     if (cache_active && fileExtension != "dds" && fileExtension != "gz" &&
         (!sgoptC
             || (sgoptC->getLoadOriginHint() != SGReaderWriterOptions::LoadOriginHint::ORIGIN_SPLASH_SCREEN
+                && sgoptC->getLoadOriginHint() != SGReaderWriterOptions::LoadOriginHint::ORIGIN_MATERIAL_ATLAS
                 && sgoptC->getLoadOriginHint() != SGReaderWriterOptions::LoadOriginHint::ORIGIN_CANVAS))) {
         std::string root = getPathRoot(absFileName);
         std::string prr = getPathRelative(root, absFileName);
@@ -558,6 +562,9 @@ ModelRegistry::readImage(const string& fileName,
 
 
     try {
+        simgear::ReportBadAllocGuard g;
+        // REVIEW: Memory Leak - 262,144 bytes in 1 blocks are indirectly lost
+        // The leak occurs with OSG, but may be related to opt being a raw pointer
         res = registry->readImageImplementation(absFileName, opt);
     } catch (std::bad_alloc&) {
         simgear::reportFailure(simgear::LoadFailure::OutOfMemory, simgear::ErrorCode::ThreeDModelLoad,
@@ -565,6 +572,7 @@ ModelRegistry::readImage(const string& fileName,
         return ReaderWriter::ReadResult::INSUFFICIENT_MEMORY_TO_LOAD;
     }
 
+
     if (!res.success()) {
         SG_LOG(SG_IO, SG_DEV_WARN, "Image loading failed:" << res.message());
         return res;
@@ -610,12 +618,13 @@ ModelRegistry::readImage(const string& fileName,
 
 osg::ref_ptr<osg::Node> DefaultCachePolicy::find(const string& fileName, const Options* opt)
 {
+    // If the options explicitly ask for no cache, do not use the cached version, even if it exists
+    if (opt->getObjectCacheHint() == osgDB::Options::CACHE_NONE) {
+        SG_LOG(SG_IO, SG_BULK, "Reading model \"" << fileName << "\"");
+        return NULL;
+    }
     Registry* registry = Registry::instance();
-#if OSG_VERSION_LESS_THAN(3,4,0)
-    osg::ref_ptr<osg::Object> cachedObject = registry->getFromObjectCache(fileName);
-#else
     osg::ref_ptr<osg::Object> cachedObject = registry->getRefFromObjectCache(fileName);
-#endif
 
     ref_ptr<osg::Node> cachedNode = dynamic_cast<osg::Node*>(cachedObject.get());
     if (cachedNode.valid())
@@ -642,31 +651,44 @@ void DefaultCachePolicy::addToCache(const string& fileName,
 // opts |= osgUtil::Optimizer::COPY_SHARED_NODES;
 // opts |= osgUtil::Optimizer::TESSELATE_GEOMETRY;
 // opts |= osgUtil::Optimizer::OPTIMIZE_TEXTURE_SETTINGS;
+//
+// There two significantly slow down reading of large disconnected .ac files (180k vertex - 30s)
+// for no apparent framerate benefit as modern graphics cards don't have a shared
+// vertex cache anymore:
+// opts |= osgUtil::Optimizer::VERTEX_POSTTRANSFORM
+// opts |= osgUtil::Optimizer::VERTEX_PRETRANSFORM
 
 OptimizeModelPolicy::OptimizeModelPolicy(const string& extension) :
     _osgOptions(Optimizer::SHARE_DUPLICATE_STATE
                 | Optimizer::MERGE_GEOMETRY
                 | Optimizer::FLATTEN_STATIC_TRANSFORMS
                 | Optimizer::INDEX_MESH
-                | Optimizer::VERTEX_POSTTRANSFORM
-                | Optimizer::VERTEX_PRETRANSFORM)
+    )
 {
 }
 
 osg::Node* OptimizeModelPolicy::optimize(osg::Node* node,
                                          const string& fileName,
-                                         const osgDB::Options* opt)
+                                         const osgDB::Options* opt,
+                                         const bool compressTextures)
 {
     osgUtil::Optimizer optimizer;
+
+    auto start = std::chrono::system_clock::now();
     optimizer.optimize(node, _osgOptions);
+    std::chrono::duration<double> elapsed_seconds = std::chrono::system_clock::now() - start;
+    SG_LOG(SG_IO, SG_DEBUG, "Optimization of " << fileName << " took " << elapsed_seconds.count());
 
     // Make sure the data variance of sharable objects is set to
     // STATIC so that textures will be globally shared.
     SGTexDataVarianceVisitor dataVarianceVisitor;
     node->accept(dataVarianceVisitor);
 
-    SGTexCompressionVisitor texComp;
-    node->accept(texComp);
+    if (compressTextures) {
+        SGTexCompressionVisitor texComp;
+        node->accept(texComp);
+    }
+
     return node;
 }
 
@@ -679,6 +701,32 @@ string OSGSubstitutePolicy::substitute(const string& name,
     return absFileName;
 }
 
+string ArchiveSubstitutePolicy::substitute(const string& name,
+                                          const Options* opt)
+{
+    std::size_t wsPos = name.find(ModelRegistry::WS30_PREFIX);
+    std::size_t zipPos = name.find(ModelRegistry::WS30_ARCHIVE_EXT);
+    std::size_t subdirPos = name.find(ModelRegistry::WS30_SUBDIR_SUFFIX);
+
+    if ((wsPos != string::npos) && (zipPos != string::npos) && (subdirPos != string::npos)) {
+        // .osg files that reference other .osg files have a file path relative to the directory
+        // of the parent .osg file.  However, there is an OSG bug which means this does not work
+        // in the case of archives, where any directory within the archive is lost.
+        // To work around this, WS30 creates archives without any directories within them, but
+        // any filepath reference in a zip file that includes a sub-directory needs to be 
+        // stripped out.
+        //
+        // So we need to change
+        //   vpb/w070n10/w061n14.zip/ws_w061n14_root_L0_X0_Y0/ws_w061n14_L0_X0_Y0_subtile.osgb
+        //   vpb/w070n10/w061n14.zip/ws_w061n14_L0_X0_Y0_subtile.osgb
+        std::string archivePath = name.substr(0 ,zipPos + 4);
+        archivePath.append(name.substr(subdirPos + ModelRegistry::WS30_SUBDIR_SUFFIX.length()));
+        SG_LOG(SG_IO, SG_DEBUG, "Converted path " << name << " to " << archivePath);
+        return archivePath;
+    } else {
+        return std::string();
+    }
+}
 
 void
 BuildLeafBVHPolicy::buildBVH(const std::string& fileName, osg::Node* node)
@@ -724,10 +772,24 @@ ModelRegistry::addNodeCallbackForExtension(const string& extension,
     nodeCallbackMap.insert(CallbackMap::value_type(extension, callback));
 }
 
+static bool fileNameIsFGDataModelXML(const std::string& fileName)
+{
+    if (!simgear::strutils::ends_with(fileName, ".xml")) {
+        return false;
+    }
+
+    // we could instead white-list all the renamed files here?
+    return simgear::strutils::starts_with(fileName, "Models/");
+}
+
 ReaderWriter::ReadResult
 ModelRegistry::readNode(const string& fileName,
                         const Options* opt)
 {
+    if (fileName.empty()) {
+        return osgDB::ReaderWriter::ReadResult::FILE_NOT_HANDLED;
+    }
+
     // propogate error context from the caller
     simgear::ErrorReportContext ec;
     auto sgopt = dynamic_cast<const SGReaderWriterOptions*>(opt);
@@ -739,8 +801,7 @@ ModelRegistry::readNode(const string& fileName,
         ec.addFromMap(sgopt->getErrorContext());
     }
 
-    CallbackMap::iterator iter
-        = nodeCallbackMap.find(getFileExtension(fileName));
+    auto iter = nodeCallbackMap.find(getFileExtension(fileName));
     ReaderWriter::ReadResult result;
     if (iter != nodeCallbackMap.end() && iter->second.valid())
         result = iter->second->readNode(fileName, opt);
@@ -748,6 +809,22 @@ ModelRegistry::readNode(const string& fileName,
         result = _defaultCallback->readNode(fileName, opt);
 
     if (!result.validNode()) {
+        if (fileNameIsFGDataModelXML(fileName)) {
+            const auto finalDotPos = fileName.rfind(".");
+            const auto plainModelFileName = fileName.substr(0, finalDotPos) + ".ac";
+
+            // try this AC reader; if it succeeds, we can use that.
+            auto iter = nodeCallbackMap.find("ac");
+            result = iter->second->readNode(plainModelFileName, opt);
+            if (result.validNode()) {
+                // TODO: allow extracting keys from the error-context;
+                // would need an additional callback added with a return value
+                // const auto stgName = ec.get("terrain-stg");
+                SG_LOG(SG_IO, SG_DEV_WARN, "Couldn't find XML model, found via .AC instead:" << plainModelFileName);
+                return result;
+            }
+        }
+
         simgear::reportFailure(simgear::LoadFailure::BadData, simgear::ErrorCode::ThreeDModelLoad,
                                "Failed to load 3D model:" + result.message(), sg_location{fileName});
     }
@@ -842,9 +919,11 @@ struct ACProcessPolicy {
     }
 };
 
-typedef ModelRegistryCallback<ACProcessPolicy, DefaultCachePolicy,
+typedef ModelRegistryCallback<ACProcessPolicy,
+                              DefaultCachePolicy,
                               ACOptimizePolicy,
-                              OSGSubstitutePolicy, BuildLeafBVHPolicy>
+                              OSGSubstitutePolicy,
+                              BuildLeafBVHPolicy>
 ACCallback;
 
 struct OBJProcessPolicy {
@@ -862,7 +941,8 @@ struct OBJProcessPolicy {
 typedef ModelRegistryCallback<OBJProcessPolicy,
                               DefaultCachePolicy,
                               ACOptimizePolicy,
-                              OSGSubstitutePolicy, BuildLeafBVHPolicy>
+                              OSGSubstitutePolicy,
+                              BuildLeafBVHPolicy>
 OBJCallback;
 
 
@@ -904,6 +984,122 @@ struct IVEOptimizePolicy : public OptimizeModelPolicy {
     }
 };
 
+template<class T>
+class FindTopMostNodeOfTypeVisitor : public osg::NodeVisitor
+{
+public:
+    FindTopMostNodeOfTypeVisitor():
+        osg::NodeVisitor(osg::NodeVisitor::TRAVERSE_ACTIVE_CHILDREN),
+        _foundNode(0)
+    {}
+
+    void apply(osg::Node& node)
+    {
+        T* result = dynamic_cast<T*>(&node);
+        if (result)
+        {
+            _foundNode = result;
+        }
+        else if (! dynamic_cast<osg::PagedLOD*>(&node))
+        {
+            traverse(node);
+        }
+    }
+
+    T* _foundNode;
+};
+
+template<class T>
+T* findTopMostNodeOfType(osg::Node* node)
+{
+    if (!node) return 0;
+
+    FindTopMostNodeOfTypeVisitor<T> fnotv;
+    node->accept(fnotv);
+
+    return fnotv._foundNode;
+}
+
+class CleanTechniqueVisitor : public osg::NodeVisitor
+{
+public:
+    CleanTechniqueVisitor():
+        osg::NodeVisitor(osg::NodeVisitor::TRAVERSE_ALL_CHILDREN) {}
+
+    void apply(osg::Node& node)
+    {
+        osgTerrain::TerrainTile* tile = dynamic_cast<osgTerrain::TerrainTile*>(&node);
+        if (tile)
+        {
+            if (tile->getTerrainTechnique())
+            {
+                // OSG_NOTICE<<"Resetting TerrainTechnhique "<<tile->getTerrainTechnique()->className()<<" to 0"<<std::endl;
+                tile->setTerrainTechnique(0);
+            }
+        }
+        else
+        {
+            traverse(node);
+        }
+    }
+};
+
+struct OSGOptimizePolicy : public OptimizeModelPolicy {
+
+    
+    OSGOptimizePolicy(const string& extension) :
+        OptimizeModelPolicy(extension)
+    {
+        _osgOptions &= ~Optimizer::TRISTRIP_GEOMETRY;
+    }
+    Node* optimize(Node* node, const string& fileName,
+        const Options* opt)
+    {
+        ref_ptr<Node> optimized = node;
+
+        osg::ref_ptr<simgear::SGReaderWriterOptions> sgopt = SGReaderWriterOptions::copyOrCreate(opt);
+
+        if (fileName.find(ModelRegistry::WS30_PREFIX) != string::npos) {
+            // Currently the only way we have to identify WS3.0 / VirtualPlanetBuilder files is by the filename
+
+            SGSceneFeatures* features = SGSceneFeatures::instance();
+            osg::ref_ptr<osgTerrain::Terrain> terrain = findTopMostNodeOfType<osgTerrain::Terrain>(optimized.get());
+
+            if (terrain != NULL) {
+                // Top level.  This is likely to have the default GeometryTechnique already assigned which we need to replace with our own
+                terrain->setSampleRatio(features->getVPBSampleRatio());
+                terrain->setVerticalScale(features->getVPBVerticalScale());
+                terrain->setTerrainTechniquePrototype(new VPBTechnique(sgopt, fileName));
+            }
+        } else {
+            optimized = OptimizeModelPolicy::optimize(optimized, fileName, opt, true);
+        }
+
+        Group* group = dynamic_cast<Group*>(optimized.get());
+        MatrixTransform* transform
+            = dynamic_cast<MatrixTransform*>(optimized.get());
+        if (((transform && transform->getMatrix().isIdentity()) || group)
+            && group->getName().empty()
+            && group->getNumChildren() == 1) {
+            optimized = static_cast<Node*>(group->getChild(0));
+            group = dynamic_cast<Group*>(optimized.get());
+            if (group && group->getName().empty()
+                && group->getNumChildren() == 1)
+                optimized = static_cast<Node*>(group->getChild(0));
+        }
+
+        if (sgopt && sgopt->getInstantiateMaterialEffects()) {
+            optimized = instantiateMaterialEffects(optimized.get(), sgopt);
+        }
+        else if (sgopt && sgopt->getInstantiateEffects()) {
+            optimized = instantiateEffects(optimized.get(), sgopt);
+        }
+
+        return optimized.release();
+    }
+};
+
+
 struct IVEProcessPolicy {
     IVEProcessPolicy(const string& extension) {}
     Node* process(Node* node, const string& filename,
@@ -929,15 +1125,22 @@ struct IVEProcessPolicy {
 
 typedef ModelRegistryCallback<IVEProcessPolicy, DefaultCachePolicy,
     IVEOptimizePolicy,
-    OSGSubstitutePolicy, BuildLeafBVHPolicy>
+    OSGSubstitutePolicy, 
+    BuildLeafBVHPolicy>
     IVECallback;
 
+typedef ModelRegistryCallback<IVEProcessPolicy, NoCachePolicy,
+    OSGOptimizePolicy,
+    ArchiveSubstitutePolicy, 
+    BuildLeafBVHPolicy>
+    OSGCallback;
+
 namespace
 {
 ModelRegistryCallbackProxy<ACCallback> g_acRegister("ac");
 ModelRegistryCallbackProxy<OBJCallback> g_objRegister("obj");
 ModelRegistryCallbackProxy<IVECallback> g_iveRegister("ive");
-ModelRegistryCallbackProxy<IVECallback> g_osgtRegister("osgt");
-ModelRegistryCallbackProxy<IVECallback> g_osgbRegister("osgb");
+ModelRegistryCallbackProxy<OSGCallback> g_osgtRegister("osgt");
+ModelRegistryCallbackProxy<OSGCallback> g_osgbRegister("osgb");
 
 }
