@@ -19,10 +19,11 @@
 // Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 //
 
+#include <memory>
 #ifdef HAVE_CONFIG_H
 #  include <simgear_config.h>
 #endif
-
+#include <algorithm>
 #include "ReaderWriterSTG.hxx"
 
 #include <osg/LOD>
@@ -38,20 +39,30 @@
 #include <osgDB/ReaderWriter>
 #include <osgDB/ReadFile>
 
-#include <simgear/math/sg_random.h>
-#include <simgear/math/SGGeometry.hxx>
 #include <simgear/bucket/newbucket.hxx>
+#include <simgear/debug/ErrorReportingCallback.hxx>
 #include <simgear/debug/logstream.hxx>
+#include <simgear/math/SGGeometry.hxx>
+#include <simgear/math/sg_random.hxx>
+
 #include <simgear/io/iostreams/sgstream.hxx>
+#include <simgear/scene/material/matlib.hxx>
+#include <simgear/scene/tgdb/LightBin.hxx>
+#include <simgear/scene/tgdb/ObjectInstanceBin.hxx>
+#include <simgear/scene/tgdb/SGBuildingBin.hxx>
+#include <simgear/scene/tgdb/TreeBin.hxx>
+#include <simgear/scene/tgdb/VPBRasterRenderer.hxx>
+#include <simgear/scene/tgdb/VPBLineFeatureRenderer.hxx>
+#include <simgear/scene/tgdb/VPBTechnique.hxx>
+#include <simgear/scene/tgdb/apt_signs.hxx>
+#include <simgear/scene/tgdb/obj.hxx>
 #include <simgear/scene/util/OptionsReadFileCallback.hxx>
 #include <simgear/scene/util/OsgMath.hxx>
 #include <simgear/scene/util/QuadTreeBuilder.hxx>
 #include <simgear/scene/util/RenderConstants.hxx>
 #include <simgear/scene/util/SGReaderWriterOptions.hxx>
-#include <simgear/scene/tgdb/apt_signs.hxx>
-#include <simgear/scene/tgdb/obj.hxx>
-#include <simgear/scene/material/matlib.hxx>
-#include <simgear/scene/tgdb/SGBuildingBin.hxx>
+
+#include <simgear/scene/util/SGSceneFeatures.hxx>
 
 #include "SGOceanTile.hxx"
 
@@ -62,6 +73,12 @@
 #define RAILWAY_ROUGH "OBJECT_RAILWAY_ROUGH"
 #define RAILWAY_DETAILED "OBJECT_RAILWAY_DETAILED"
 #define BUILDING_LIST "BUILDING_LIST"
+#define TREE_LIST "TREE_LIST"
+#define LINE_FEATURE_LIST "LINE_FEATURE_LIST"
+#define COASTLINE_LIST "COASTLINE_LIST"
+#define OBJECT_LIGHT "OBJECT_LIGHT"
+#define LIGHT_LIST "LIGHT_LIST"
+#define OBJECT_INSTANCED "OBJECT_INSTANCED"
 
 namespace simgear {
 
@@ -110,7 +127,7 @@ struct ReaderWriterSTG::_ModelBin {
         osg::ref_ptr<SGReaderWriterOptions> _options;
     };
     struct _ObjectStatic {
-        _ObjectStatic() : _agl(false), _proxy(false), _lon(0), _lat(0), _elev(0), _hdg(0), _pitch(0), _roll(0), _range(SG_OBJECT_RANGE_ROUGH) { }
+        _ObjectStatic() : _agl(false), _proxy(false), _lon(0), _lat(0), _elev(0), _hdg(0), _pitch(0), _roll(0), _range(SG_OBJECT_RANGE_ROUGH), _radius(10) { }
         SGPath _errorLocation;
         std::string _token;
         std::string _name;
@@ -118,7 +135,7 @@ struct ReaderWriterSTG::_ModelBin {
         bool _proxy;
         double _lon, _lat, _elev;
         double _hdg, _pitch, _roll;
-        double _range;
+        double _range, _radius;
         osg::ref_ptr<SGReaderWriterOptions> _options;
     };
     struct _Sign {
@@ -137,6 +154,50 @@ struct ReaderWriterSTG::_ModelBin {
       std::string _material_name;
       double _lon, _lat, _elev;
     };
+    struct _TreeList {
+      _TreeList() : _lon(0), _lat(0), _elev(0) { }
+      std::string _filename;
+      std::string _material_name;
+      double _lon, _lat, _elev;
+    };
+    struct _Light {
+        _Light() : _lon(0), _lat(0), _elev(0), _size(0), _intensity(0), _on_period(0), _horizontal_angle(0), _vertical_angle(0) { }
+        double _lon, _lat, _elev;
+        double _size, _intensity;
+        int _on_period;
+        SGVec4f _color;
+        SGVec3f _direction;
+        double _horizontal_angle, _vertical_angle;
+        SGVec4f _animation_params;
+    };
+    struct _LightList {
+      _LightList() : _lon(0), _lat(0), _elev(0) { }
+      std::string _filename;
+      double _lon, _lat, _elev;
+    };
+
+    struct _InstancedObject {
+        _InstancedObject() : _lon(0), _lat(0), _elev(0) {}
+        SGPath _errorLocation;
+        SGPath _STGFilePath;
+        std::string _modelname;
+        std::string _filename;
+        std::string _effect;
+        double _lon, _lat, _elev;
+        osg::ref_ptr<SGReaderWriterOptions> _options;
+    };
+
+    struct _LineFeatureList {
+      _LineFeatureList() { }
+      std::string _filename;
+      std::string _material;
+      SGBucket _bucket;
+    };
+    struct _CoastlineList {
+      _CoastlineList() { }
+      std::string _filename;
+      SGBucket _bucket;
+    };
 
     class DelayLoadReadFileCallback : public OptionsReadFileCallback {
 
@@ -154,20 +215,17 @@ struct ReaderWriterSTG::_ModelBin {
                 proxy->setName("proxyNode");
                 proxy->setLoadingExternalReferenceMode(osg::ProxyNode::DEFER_LOADING_TO_DATABASE_PAGER);
                 proxy->setFileName(0, o._name);
-                proxy->setDatabaseOptions(o._options.get());
+                proxy->setDatabaseOptions(o._options);
 
                 // Give the node some values so the Quadtree builder has
                 // a BoundingBox to work with prior to the model being loaded.
                 proxy->setCenter(osg::Vec3f(0.0f,0.0f,0.0f));
-                proxy->setRadius(10);
+                proxy->setRadius(o._radius);
                 proxy->setCenterMode(osg::ProxyNode::UNION_OF_BOUNDING_SPHERE_AND_USER_DEFINED);
                 node = proxy;
             } else {
-#if OSG_VERSION_LESS_THAN(3,4,0)
-                node = osgDB::readNodeFile(o._name, o._options.get());
-#else
+                ErrorReportContext ec("terrain-stg", o._errorLocation.utf8Str());
                 node = osgDB::readRefNodeFile(o._name, o._options.get());
-#endif
                 if (!node.valid()) {
                     SG_LOG(SG_TERRAIN, SG_ALERT, o._errorLocation << ": Failed to load "
                            << o._token << " '" << o._name << "'");
@@ -209,6 +267,8 @@ struct ReaderWriterSTG::_ModelBin {
         virtual osgDB::ReaderWriter::ReadResult
         readNode(const std::string&, const osgDB::Options*)
         {
+            ErrorReportContext ec("terrain-bucket", _bucket.gen_index_str());
+
             STGObjectsQuadtree quadtree((GetModelLODCoord()), (AddModelLOD()));
             quadtree.buildQuadTree(_objectStaticList.begin(), _objectStaticList.end());
             osg::ref_ptr<osg::Group> group = quadtree.getRoot();
@@ -242,7 +302,7 @@ struct ReaderWriterSTG::_ModelBin {
                         }
 
                         const auto path = SGPath(b._filename);
-                        SGBuildingBin* buildingBin = new SGBuildingBin(path, mat, useVBOs);
+                        auto buildingBin = std::make_shared<SGBuildingBin>(path, mat, useVBOs);
 
                         SGBuildingBinList bbList;
                         bbList.push_back(buildingBin);
@@ -253,10 +313,108 @@ struct ReaderWriterSTG::_ModelBin {
                         matrixTransform->setDataVariance(osg::Object::STATIC);
                         matrixTransform->addChild(createRandomBuildings(bbList, osg::Matrix::identity(), _options));
                         group->addChild(matrixTransform);
+                    }
+                }
+            }
 
-                        std::for_each(bbList.begin(), bbList.end(), [](SGBuildingBin* bb) {
+            if (!_treeList.empty()) {
+                SGMaterialLibPtr matlib = _options->getMaterialLib();
+
+                if (!matlib) {
+                    SG_LOG( SG_TERRAIN, SG_ALERT, "Unable to get materials definition for buildings");
+                } else {
+                    for (const auto& b : _treeList) {
+                        // Build trees for each list of trees
+                        SGGeod geodPos = SGGeod::fromDegM(b._lon, b._lat, 0.0);
+                        SGSharedPtr<SGMaterial> mat = matlib->find(b._material_name, geodPos);
+
+                        // trying to avoid crash on null material, see:
+                        // https://sentry.io/organizations/flightgear/issues/1867075869
+                        if (!mat) {
+                            SG_LOG(SG_TERRAIN, SG_ALERT, "Tree list specifies unknown material: " << b._filename << " " << b._material_name);
+                            continue;
+                        }
+
+                        const auto path = SGPath(b._filename);
+                        TreeBin* treeBin = new TreeBin(path, mat);
+
+                        SGTreeBinList treeList;
+                        treeList.push_back(treeBin);
+
+                        osg::MatrixTransform* matrixTransform;
+                        matrixTransform = new osg::MatrixTransform(makeZUpFrame(SGGeod::fromDegM(b._lon, b._lat, b._elev)));
+                        matrixTransform->setName("rotateTrees");
+                        matrixTransform->setDataVariance(osg::Object::STATIC);
+                        matrixTransform->addChild(createForest(treeList, osg::Matrix::identity(), _options));
+                        group->addChild(matrixTransform);
+
+                        std::for_each(treeList.begin(), treeList.end(), [](TreeBin* bb) {
+                            // We're finished with the intermediate data structure, so just delete it.
                             delete bb;
                         });
+
+                        treeList.clear();
+                    }
+                }
+            }
+
+            // Add lights
+            if (!_lightList.empty()) {
+                // Transform lights frame of ref for better precision
+                osg::MatrixTransform* matrixTransform;
+                matrixTransform = new osg::MatrixTransform(makeZUpFrame(_bucket.get_center()));
+                matrixTransform->setName("rotateLights");
+                matrixTransform->setDataVariance(osg::Object::STATIC);
+
+                LightBin lightList;
+                for (const auto& light : _lightList) {
+                    osg::Matrix _position_frame(makeZUpFrame(SGGeod::fromDegM(light._lon, light._lat, light._elev)));
+                    SGVec3f _position(_position_frame(3,0), _position_frame(3,1), _position_frame(3,2));
+
+                    lightList.insert(
+                        _position,
+                        light._size, light._intensity,
+                        light._on_period,
+                        light._color,
+                        light._direction,
+                        light._horizontal_angle, light._vertical_angle,
+                        light._animation_params
+                    );
+                }
+
+                matrixTransform->addChild(createLights(lightList, matrixTransform->getInverseMatrix(), _options));
+                group->addChild(matrixTransform);
+            }
+
+            if (!_lightListList.empty()) {
+                for (const auto& ll : _lightListList) {
+                    osg::MatrixTransform* matrixTransform;
+                    matrixTransform = new osg::MatrixTransform(makeZUpFrame(SGGeod::fromDegM(ll._lon, ll._lat, ll._elev)));
+                    matrixTransform->setName("rotateLights");
+                    matrixTransform->setDataVariance(osg::Object::STATIC);
+
+                    const auto path = SGPath(ll._filename);
+                    LightBin lightList(path);
+
+                    matrixTransform->addChild(createLights(lightList, osg::Matrix::identity(), _options));
+                    group->addChild(matrixTransform);
+                }
+            }
+
+            if (!_instancedObjectList.empty()) {
+                for (const auto& io : _instancedObjectList) {
+                    osg::MatrixTransform* matrixTransform;
+                    matrixTransform = new osg::MatrixTransform(makeZUpFrame(SGGeod::fromDegM(io._lon, io._lat, io._elev)));
+                    matrixTransform->setName("rotateInstancedObject");
+                    matrixTransform->setDataVariance(osg::Object::STATIC);
+
+                    const auto path = SGPath(io._filename);
+                    ObjectInstanceBin objectInstances(io._modelname, io._effect, io._STGFilePath, path);
+
+                    const auto instancedNode = createObjectInstances(objectInstances, osg::Matrix::identity(), io._options);
+                    if (instancedNode) {
+                        matrixTransform->addChild(instancedNode);
+                        group->addChild(matrixTransform);
                     }
                 }
             }
@@ -268,6 +426,10 @@ struct ReaderWriterSTG::_ModelBin {
         std::list<_ObjectStatic> _objectStaticList;
         std::list<_Sign> _signList;
         std::list<_BuildingList> _buildingList;
+        std::list<_TreeList> _treeList;
+        std::list<_Light> _lightList;
+        std::list<_LightList> _lightListList;
+        std::list<_InstancedObject> _instancedObjectList;
 
         /// The original options to use for this bunch of models
         osg::ref_ptr<SGReaderWriterOptions> _options;
@@ -378,8 +540,7 @@ struct ReaderWriterSTG::_ModelBin {
         double bareRangeDelta = atof(options->getPluginStringData("SimGear::LOD_RANGE_BARE").c_str());
         double roughRangeDelta = atof(options->getPluginStringData("SimGear::LOD_RANGE_ROUGH").c_str());
 
-        // Determine object ranges.  Mesh size of 2000mx2000m needs to be accounted for.
-        _object_range_detailed = 1414.0f + detailedRange;
+        _object_range_detailed = detailedRange;
         _object_range_bare = _object_range_detailed + bareRangeDelta;
         _object_range_rough = _object_range_detailed + roughRangeDelta;
 
@@ -389,9 +550,9 @@ struct ReaderWriterSTG::_ModelBin {
 
         // Bucket provides a consistent seed
         // so we have consistent set of pseudo-random numbers for each STG file
-        mt seed;
-        int bucket = atoi(absoluteFileName.file_base().c_str());
-        mt_init(&seed, bucket);
+        pc_init(std::stoi(absoluteFileName.file_base()));
+
+        bool vpb_active = SGSceneFeatures::instance()->getVPBActive();
 
         // do only load airport btg files.
         bool onlyAirports = options->getPluginStringData("SimGear::FG_ONLY_AIRPORTS") == "ON";
@@ -426,18 +587,19 @@ struct ReaderWriterSTG::_ModelBin {
             path.append(name);
 
             if (token == "OBJECT_BASE") {
-                // Load only once (first found)
-                SG_LOG( SG_TERRAIN, SG_BULK, "    " << token << " " << name );
-                _foundBase = true;
-                if (!onlyAirports || isAirportBtg(name)) {
-                    _Object obj;
-                    obj._errorLocation = absoluteFileName;
-                    obj._token = token;
-                    obj._name = path.utf8Str();
-                    obj._options = staticOptions(filePath, options);
-                    _objectList.push_back(obj);
+                if (!vpb_active) {
+                    // Load only once (first found)
+                    SG_LOG( SG_TERRAIN, SG_BULK, "    " << token << " " << name );
+                    _foundBase = true;
+                    if (!onlyAirports || isAirportBtg(name)) {
+                        _Object obj;
+                        obj._errorLocation = absoluteFileName;
+                        obj._token = token;
+                        obj._name = path.utf8Str();
+                        obj._options = staticOptions(filePath, options);
+                        _objectList.push_back(obj);
+                    }
                 }
-
             } else if (token == "OBJECT") {
                 if (!onlyAirports || isAirportBtg(name)) {
                     _Object obj;
@@ -447,60 +609,60 @@ struct ReaderWriterSTG::_ModelBin {
                     obj._options = staticOptions(filePath, options);
                     _objectList.push_back(obj);
                 }
-
             } else if (!onlyTerrain) {
                 // Load non-terrain objects
 
-								// Determine an appropriate range for the object, which has some randomness
-								double range = _object_range_rough;
-								double lrand = mt_rand(&seed);
-								if      (lrand < 0.1) range = range * 2.0;
-								else if (lrand < 0.4) range = range * 1.5;
+                // Determine an appropriate range for the object, which has some randomness
+                double range = _object_range_rough;
+                double lrand = pc_rand();
+                if      (lrand < 0.1) range = range * 2.0;
+                else if (lrand < 0.4) range = range * 1.5;
 
                 if (token == "OBJECT_STATIC" || token == "OBJECT_STATIC_AGL") {
-										osg::ref_ptr<SGReaderWriterOptions> opt;
-										opt = staticOptions(filePath, options);
-										if (SGPath(name).lower_extension() == "ac")
-												opt->setInstantiateEffects(true);
-										else
-												opt->setInstantiateEffects(false);
-										_ObjectStatic obj;
+                    osg::ref_ptr<SGReaderWriterOptions> opt;
+                    opt = staticOptions(filePath, options);
+                    if (SGPath(name).lower_extension() == "ac")
+                            opt->setInstantiateEffects(true);
+                    else
+                            opt->setInstantiateEffects(false);
+                    _ObjectStatic obj;
 
-										obj._errorLocation = absoluteFileName;
-										obj._token = token;
-										obj._name = name;
-										obj._agl = (token == "OBJECT_STATIC_AGL");
-										obj._proxy = true;
-										in >> obj._lon >> obj._lat >> obj._elev >> obj._hdg >> obj._pitch >> obj._roll;
-										obj._range = range;
-										obj._options = opt;
+                    opt->addErrorContext("terrain-stg", absoluteFileName.utf8Str());
+                    obj._errorLocation = absoluteFileName;
+                    obj._token = token;
+                    obj._name = name;
+                    obj._agl = (token == "OBJECT_STATIC_AGL");
+                    obj._proxy = true;
+                    in >> obj._lon >> obj._lat >> obj._elev >> obj._hdg >> obj._pitch >> obj._roll >> obj._radius;
+                    obj._range = range;
+                    obj._options = opt;
                     checkInsideBucket(absoluteFileName, obj._lon, obj._lat);
                     _objectStaticList.push_back(obj);
                 } else if (token == "OBJECT_SHARED" || token == "OBJECT_SHARED_AGL") {
-										osg::ref_ptr<SGReaderWriterOptions> opt;
-										opt = sharedOptions(filePath, options);
-										if (SGPath(name).lower_extension() == "ac")
-												opt->setInstantiateEffects(true);
-										else
-												opt->setInstantiateEffects(false);
-										_ObjectStatic obj;
-										obj._errorLocation = absoluteFileName;
-										obj._token = token;
-										obj._name = name;
-										obj._agl = (token == "OBJECT_SHARED_AGL");
-										obj._proxy = false;
-										in >> obj._lon >> obj._lat >> obj._elev >> obj._hdg >> obj._pitch >> obj._roll;
-										obj._range = range;
-										obj._options = opt;
+                    osg::ref_ptr<SGReaderWriterOptions> opt;
+                    opt = sharedOptions(filePath, options);
+                    if (SGPath(name).lower_extension() == "ac")
+                            opt->setInstantiateEffects(true);
+                    else
+                            opt->setInstantiateEffects(false);
+                    _ObjectStatic obj;
+                    obj._errorLocation = absoluteFileName;
+                    obj._token = token;
+                    obj._name = name;
+                    obj._agl = (token == "OBJECT_SHARED_AGL");
+                    obj._proxy = false;
+                    in >> obj._lon >> obj._lat >> obj._elev >> obj._hdg >> obj._pitch >> obj._roll >> obj._radius;
+                    obj._range = range;
+                    obj._options = opt;
                     checkInsideBucket(absoluteFileName, obj._lon, obj._lat);
                     _objectStaticList.push_back(obj);
                 } else if (token == "OBJECT_SIGN" || token == "OBJECT_SIGN_AGL") {
-										_Sign sign;
-										sign._token = token;
-										sign._name = name;
-										sign._agl = (token == "OBJECT_SIGN_AGL");
-										in >> sign._lon >> sign._lat >> sign._elev >> sign._hdg >> sign._size;
-										_signList.push_back(sign);
+                    _Sign sign;
+                    sign._token = token;
+                    sign._name = name;
+                    sign._agl = (token == "OBJECT_SIGN_AGL");
+                    in >> sign._lon >> sign._lat >> sign._elev >> sign._hdg >> sign._size;
+                    _signList.push_back(sign);
                 } else if (token == BUILDING_ROUGH || token == BUILDING_DETAILED ||
                            token == ROAD_ROUGH     || token == ROAD_DETAILED     ||
                            token == RAILWAY_ROUGH  || token == RAILWAY_DETAILED)   {
@@ -509,6 +671,8 @@ struct ReaderWriterSTG::_ModelBin {
                     _ObjectStatic obj;
 
                     opt->setInstantiateEffects(false);
+                    opt->addErrorContext("terrain-stg", absoluteFileName.utf8Str());
+
                     if (SGPath(name).lower_extension() == "ac") {
                       // Generate material/Effects lookups for raw models, i.e.
                       // those not wrapped in an XML while will include Effects
@@ -532,17 +696,17 @@ struct ReaderWriterSTG::_ModelBin {
                     obj._name = name;
                     obj._agl = false;
                     obj._proxy = true;
-                    in >> obj._lon >> obj._lat >> obj._elev >> obj._hdg >> obj._pitch >> obj._roll;
+                    in >> obj._lon >> obj._lat >> obj._elev >> obj._hdg >> obj._pitch >> obj._roll >> obj._radius;
 
                     opt->setLocation(obj._lon, obj._lat);
                     if (token == BUILDING_DETAILED || token == ROAD_DETAILED || token == RAILWAY_DETAILED ) {
-											// Apply a lower LOD range if this is a detailed mesh
-											range = _object_range_detailed;
-											double lrand = mt_rand(&seed);
-											if      (lrand < 0.1) range = range * 2.0;
-											else if (lrand < 0.4) range = range * 1.5;
-										}
-
+                        // Apply a lower LOD range if this is a detailed mesh.
+                        range = _object_range_detailed;
+                        double lrand = pc_rand();
+                        if      (lrand < 0.1) range = range * 2.0;
+                        else if (lrand < 0.4) range = range * 1.5;                                        
+                    }
+                    
                     obj._range = range;
 
                     obj._options = opt;
@@ -554,7 +718,60 @@ struct ReaderWriterSTG::_ModelBin {
                   in >> buildinglist._material_name >> buildinglist._lon >> buildinglist._lat >> buildinglist._elev;
                   checkInsideBucket(absoluteFileName, buildinglist._lon, buildinglist._lat);
                   _buildingListList.push_back(buildinglist);
-                  //SG_LOG(SG_TERRAIN, SG_ALERT, "Building list: " << buildinglist._filename << " " << buildinglist._material_name << " " << buildinglist._lon << " " << buildinglist._lat);
+                } else if (token == TREE_LIST) {
+                  _TreeList treelist;
+                  treelist._filename = path.utf8Str();
+                  in >> treelist._material_name >> treelist._lon >> treelist._lat >> treelist._elev;
+                  checkInsideBucket(absoluteFileName, treelist._lon, treelist._lat);
+                  _treeListList.push_back(treelist);
+                } else if (token == LINE_FEATURE_LIST) {
+                  _LineFeatureList lineFeaturelist;
+                  lineFeaturelist._filename = path.utf8Str();
+                  in >> lineFeaturelist._material;
+                  lineFeaturelist._bucket = bucketIndexFromFileName(absoluteFileName.file_base().c_str());
+                  _lineFeatureListList.push_back(lineFeaturelist);
+                } else if (token == COASTLINE_LIST) {
+                  _CoastlineList coastFeaturelist;
+                  coastFeaturelist._filename = path.utf8Str();
+                  coastFeaturelist._bucket = bucketIndexFromFileName(absoluteFileName.file_base().c_str());
+                  _coastFeatureListList.push_back(coastFeaturelist);
+                } else if (token == OBJECT_LIGHT) {
+                    _Light light;
+                    in >> light._lon >> light._lat >> light._elev
+                        >> light._size >> light._intensity
+                        >> light._on_period
+                        >> light._color[0] >> light._color[1] >> light._color[2] >> light._color[3]
+                        >> light._direction[0] >> light._direction[1] >> light._direction[2]
+                        >> light._horizontal_angle >> light._vertical_angle
+                        >> light._animation_params[0] >> light._animation_params[1] >> light._animation_params[2] >> light._animation_params[3];
+                    checkInsideBucket(absoluteFileName, light._lon, light._lat);
+                    _lightList.push_back(light);
+                } else if (token == LIGHT_LIST) {
+                    _LightList lightList;
+                    lightList._filename = path.utf8Str();
+                    in >> lightList._lon >> lightList._lat >> lightList._elev;
+                    checkInsideBucket(absoluteFileName, lightList._lon, lightList._lat);
+                    _lightListList.push_back(lightList);
+                } else if (token == OBJECT_INSTANCED) {
+                    _InstancedObject instancedObject;
+                    instancedObject._modelname = name;
+
+                    std::string filename;
+                    in >> filename;
+                    SGPath _filepath = filePath;
+                    _filepath.append(filename);
+
+                    instancedObject._errorLocation = absoluteFileName;
+                    instancedObject._STGFilePath = absoluteFileName;
+                    instancedObject._filename = _filepath.utf8Str();
+                    instancedObject._options = SGReaderWriterOptions::copyOrCreate(options);
+                    instancedObject._options->addErrorContext("terrain-stg", absoluteFileName.utf8Str());
+
+                    in >> instancedObject._effect;
+                    in >> instancedObject._lon >> instancedObject._lat >> instancedObject._elev;
+                    checkInsideBucket(absoluteFileName, instancedObject._lon, instancedObject._lat);
+
+                    _instancedObjectList.push_back(instancedObject);
                 } else {
                     // Check registered callback for token. Keep lock until callback completed to make sure it will not be
                     // executed after a thread successfully executed removeSTGObjectHandler()
@@ -573,7 +790,9 @@ struct ReaderWriterSTG::_ModelBin {
                             }
                             callback(token,name, SGGeod::fromDegM(obj._lon, obj._lat, obj._elev), obj._hdg,restofline);
                         } else {
-                            SG_LOG( SG_TERRAIN, SG_ALERT, absoluteFileName << ": Unknown token '" << token << "'" );
+                            // SG_LOG( SG_TERRAIN, SG_ALERT, absoluteFileName << ": Unknown token '" << token << "'" );
+                            simgear::reportFailure(simgear::LoadFailure::Misconfigured, simgear::ErrorCode::BTGLoad,
+                                                   "Unknown STG token:" + token, absoluteFileName);
                         }
                     }
                 }
@@ -594,17 +813,64 @@ struct ReaderWriterSTG::_ModelBin {
         std::string terrain_name = string("terrain ").append(bucket.gen_index_str());
         terrainGroup->setName(terrain_name);
 
-        if (_foundBase) {
+        simgear::ErrorReportContext ec{"terrain-bucket", bucket.gen_index_str()};
+
+        bool vpb_active = SGSceneFeatures::instance()->getVPBActive();
+        if (vpb_active) {
+
+            // Load any line area or coastline features, which we will need before we generate the tile.
+            if (!_lineFeatureListList.empty()) {
+
+                LineFeatureBinList lineFeatures;
+
+                for (const auto& b : _lineFeatureListList) {
+                    // add the lineFeatures to the list
+                    const auto path = SGPath(b._filename);
+                    lineFeatures.push_back(new LineFeatureBin(path, b._material));
+                }
+
+                VPBLineFeatureRenderer::addLineFeatureList(bucket, lineFeatures);
+            }
+
+            if (!_coastFeatureListList.empty()) {
+
+                CoastlineBinList coastFeatures;
+
+                for (const auto& b : _coastFeatureListList) {
+                    // add the lineFeatures to the list
+                    const auto path = SGPath(b._filename);
+                    coastFeatures.push_back(new CoastlineBin(path));
+                }
+
+                VPBRasterRenderer::addCoastlineList(bucket, coastFeatures);
+            }
+
+            // OBJECTs include airports
             for (auto stgObject : _objectList) {
                 osg::ref_ptr<osg::Node> node;
-#if OSG_VERSION_LESS_THAN(3,4,0)
-                node = osgDB::readNodeFile(stgObject._name, stgObject._options.get());
-#else
                 node = osgDB::readRefNodeFile(stgObject._name, stgObject._options.get());
-#endif
+
                 if (!node.valid()) {
                     SG_LOG(SG_TERRAIN, SG_ALERT, stgObject._errorLocation << ": Failed to load "
-                           << stgObject._token << " '" << stgObject._name << "'");
+                        << stgObject._token << " '" << stgObject._name << "'");
+                    continue;
+                }
+
+                // Add the OBJECT to the elevation constraints of the terrain so the terrain
+                // doesn't poke through the airport
+                simgear::VPBTechnique::addElevationConstraint(node);
+
+                terrainGroup->addChild(node.get());
+            }
+        } else if (_foundBase) {
+            for (auto stgObject : _objectList) {
+                osg::ref_ptr<osg::Node> node;
+                simgear::ErrorReportContext ec("terrain-stg", stgObject._errorLocation.utf8Str());
+                node = osgDB::readRefNodeFile(stgObject._name, stgObject._options.get());
+
+                if (!node.valid()) {
+                    SG_LOG(SG_TERRAIN, SG_ALERT, stgObject._errorLocation << ": Failed to load "
+                        << stgObject._token << " '" << stgObject._name << "'");
                     continue;
                 }
                 terrainGroup->addChild(node.get());
@@ -634,42 +900,59 @@ struct ReaderWriterSTG::_ModelBin {
             i->_elev += elevation(*terrainGroup, SGGeod::fromDeg(i->_lon, i->_lat));
         }
 
-        if (_objectStaticList.empty() && _signList.empty() && (_buildingListList.size() == 0)) {
+        if (_objectStaticList.empty() &&
+            _signList.empty() &&
+            _buildingListList.empty() &&
+            _treeListList.empty() &&
+            _lineFeatureListList.empty() &&
+            _coastFeatureListList.empty() &&
+            _lightList.empty() &&
+            _lightListList.empty() &&
+            _instancedObjectList.empty()) {
             // The simple case, just return the terrain group
             return terrainGroup.release();
-        } else {
-            osg::PagedLOD* pagedLOD = new osg::PagedLOD;
-            pagedLOD->setCenterMode(osg::PagedLOD::USE_BOUNDING_SPHERE_CENTER);
-            std::string name = string("pagedObjectLOD ").append(bucket.gen_index_str());
-            pagedLOD->setName(name);
-
-            // This should be visible in any case.
-            // If this is replaced by some lower level of detail, the parent LOD node handles this.
-            pagedLOD->addChild(terrainGroup, 0, std::numeric_limits<float>::max());
-            pagedLOD->setMinimumExpiryTime(0, pagedLODExpiry);
-
-            // we just need to know about the read file callback that itself holds the data
-            osg::ref_ptr<DelayLoadReadFileCallback> readFileCallback = new DelayLoadReadFileCallback;
-            readFileCallback->_objectStaticList = _objectStaticList;
-            readFileCallback->_buildingList = _buildingListList;
-            readFileCallback->_signList = _signList;
-            readFileCallback->_options = options;
-            readFileCallback->_bucket = bucket;
-            osg::ref_ptr<osgDB::Options> callbackOptions = new osgDB::Options;
-            callbackOptions->setReadFileCallback(readFileCallback.get());
-            pagedLOD->setDatabaseOptions(callbackOptions.get());
-
-            pagedLOD->setFileName(pagedLOD->getNumChildren(), "Dummy name - use the stored data in the read file callback");
-
-            // Objects may end up displayed up to 2x the object range.
-            pagedLOD->setRange(pagedLOD->getNumChildren(), 0, 2.0 * _object_range_rough);
-            pagedLOD->setMinimumExpiryTime(pagedLOD->getNumChildren(), pagedLODExpiry);
-            pagedLOD->setRadius(SG_TILE_RADIUS);
-            SG_LOG( SG_TERRAIN, SG_DEBUG, "Tile " << bucket.gen_index_str() << " PagedLOD Center: " << pagedLOD->getCenter().x() << "," << pagedLOD->getCenter().y() << "," << pagedLOD->getCenter().z() );
-            SG_LOG( SG_TERRAIN, SG_DEBUG, "Tile " << bucket.gen_index_str() << " PagedLOD Range: " << (2.0 * _object_range_rough));
-            SG_LOG( SG_TERRAIN, SG_DEBUG, "Tile " << bucket.gen_index_str() << " PagedLOD Radius: " << SG_TILE_RADIUS);
-            return pagedLOD;
         }
+
+        osg::PagedLOD* pagedLOD = new osg::PagedLOD;
+        std::string name = string("pagedObjectLOD ").append(bucket.gen_index_str());
+        pagedLOD->setName(name);
+
+        // This should be visible in any case.
+        // If this is replaced by some lower level of detail, the parent LOD node handles this.
+        pagedLOD->addChild(terrainGroup, 0, std::numeric_limits<float>::max());
+        pagedLOD->setMinimumExpiryTime(0, pagedLODExpiry);
+
+        // we just need to know about the read file callback that itself holds the data
+        osg::ref_ptr<DelayLoadReadFileCallback> readFileCallback = new DelayLoadReadFileCallback;
+        readFileCallback->_objectStaticList = _objectStaticList;
+        readFileCallback->_buildingList = _buildingListList;
+        readFileCallback->_treeList = _treeListList;
+        readFileCallback->_lightList = _lightList;
+        readFileCallback->_lightListList = _lightListList;
+        readFileCallback->_instancedObjectList = _instancedObjectList;
+        readFileCallback->_signList = _signList;
+        readFileCallback->_options = options;
+        readFileCallback->_bucket = bucket;
+
+        osg::ref_ptr<osgDB::Options> callbackOptions = new osgDB::Options;
+        callbackOptions->setReadFileCallback(readFileCallback.get());
+        pagedLOD->setDatabaseOptions(callbackOptions.get());
+
+        pagedLOD->setFileName(pagedLOD->getNumChildren(), "Dummy name - use the stored data in the read file callback");
+
+        // Objects may end up displayed up to 2x the object range.
+        pagedLOD->setRange(pagedLOD->getNumChildren(), 0, 2.0 * _object_range_rough);
+        pagedLOD->setMinimumExpiryTime(pagedLOD->getNumChildren(), pagedLODExpiry);
+        pagedLOD->setRadius(SG_TILE_RADIUS);
+        pagedLOD->setCenterMode(osg::PagedLOD::USER_DEFINED_CENTER);
+
+        SGVec3d coord;
+        SGGeodesy::SGGeodToCart(bucket.get_center(), coord);
+        pagedLOD->setCenter(toOsg(coord));
+        SG_LOG( SG_TERRAIN, SG_DEBUG, "Tile " << bucket.gen_index_str() << " PagedLOD Center: " << pagedLOD->getCenter().x() << "," << pagedLOD->getCenter().y() << "," << pagedLOD->getCenter().z() );
+        SG_LOG( SG_TERRAIN, SG_DEBUG, "Tile " << bucket.gen_index_str() << " PagedLOD Range: " << (2.0 * _object_range_rough));
+        SG_LOG( SG_TERRAIN, SG_DEBUG, "Tile " << bucket.gen_index_str() << " PagedLOD Radius: " << SG_TILE_RADIUS);
+        return pagedLOD;
     }
 
     double _object_range_bare;
@@ -680,6 +963,12 @@ struct ReaderWriterSTG::_ModelBin {
     std::list<_ObjectStatic> _objectStaticList;
     std::list<_Sign> _signList;
     std::list<_BuildingList> _buildingListList;
+    std::list<_TreeList> _treeListList;
+    std::list<_LineFeatureList> _lineFeatureListList;
+    std::list<_CoastlineList> _coastFeatureListList;
+    std::list<_Light> _lightList;
+    std::list<_LightList> _lightListList;
+    std::list<_InstancedObject> _instancedObjectList;
 };
 
 ReaderWriterSTG::ReaderWriterSTG()
@@ -701,6 +990,7 @@ ReaderWriterSTG::readNode(const std::string& fileName, const osgDB::Options* opt
 {
     _ModelBin modelBin;
     SGBucket bucket(bucketIndexFromFileName(fileName));
+    simgear::ErrorReportContext ec("terrain-bucket", bucket.gen_index_str());
 
     // We treat 123.stg different than ./123.stg.
     // The difference is that ./123.stg as well as any absolute path
@@ -708,6 +998,7 @@ ReaderWriterSTG::readNode(const std::string& fileName, const osgDB::Options* opt
     // In contrast 123.stg uses the search paths to load a set of stg
     // files spread across the scenery directories.
     if (osgDB::getSimpleFileName(fileName) != fileName) {
+        simgear::ErrorReportContext ec("terrain-stg", fileName);
         if (!modelBin.read(fileName, options))
             return ReadResult::FILE_NOT_FOUND;
     }
@@ -717,11 +1008,10 @@ ReaderWriterSTG::readNode(const std::string& fileName, const osgDB::Options* opt
         return ReadResult::FILE_NOT_FOUND;
     }
 
-    osg::ref_ptr<SGReaderWriterOptions> sgOpts(SGReaderWriterOptions::copyOrCreate(options));
-    if (sgOpts->getSceneryPathSuffixes().empty()) {
+    const auto sgOpts = dynamic_cast<const SGReaderWriterOptions*>(options);
+    if (!sgOpts || sgOpts->getSceneryPathSuffixes().empty()) {
         SG_LOG(SG_TERRAIN, SG_ALERT, "Loading tile " << fileName << ", no scenery path suffixes were configured so giving up");
         return ReadResult::FILE_NOT_FOUND;
-
     }
 
     SG_LOG(SG_TERRAIN, SG_INFO, "Loading tile " << fileName);
@@ -732,6 +1022,8 @@ ReaderWriterSTG::readNode(const std::string& fileName, const osgDB::Options* opt
     // But do load all STGs at the same level (i.e from the same scenery path)
     const osgDB::FilePathList& filePathList = options->getDatabasePathList();
     for (auto path : filePathList) {
+        SG_LOG(SG_TERRAIN, SG_DEBUG, "OSG DatabasePath: " << path);
+
         if (modelBin._foundBase) {
             break;
         }
@@ -746,7 +1038,8 @@ ReaderWriterSTG::readNode(const std::string& fileName, const osgDB::Options* opt
         }
 
         for (auto suffix : sgOpts->getSceneryPathSuffixes()) {
-            SGPath p = base / suffix / basePath / fileName;
+            const auto p = base / suffix / basePath / fileName;
+            simgear::ErrorReportContext ec("terrain-stg", p.utf8Str());
             modelBin.read(p, options);
         }
     }
