@@ -2,28 +2,14 @@
 //
 // Started by Curtis Olson, November 2002.
 //
-// Copyright (C) 2002  Curtis L. Olson  - http://www.flightgear.org/~curt
-// Copyright (C) 2008  Alexander R. Perry <alex.perry@ieee.org>
-// Copyright (C) 2011  Thorsten Brehm <brehmt@gmail.com>
-//
-// This program is free software; you can redistribute it and/or
-// modify it under the terms of the GNU General Public License as
-// published by the Free Software Foundation; either version 2 of the
-// License, or (at your option) any later version.
-//
-// This program is distributed in the hope that it will be useful, but
-// WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// General Public License for more details.
-//
-// You should have received a copy of the GNU General Public License
-// along with this program; if not, write to the Free Software
-// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-//
-// $Id$
+// SPDX-License-Identifier: LGPL-2.1-or-later
+// SPDX-FileCopyrightText: 2002 Curtis L. Olson
+// SPDX-FileCopyrightText: 2008 Alexander R. Perry <alex.perry@ieee.org>
+// SPDX-FileCopyrightText: 2011 Thorsten Brehm <brehmt@gmail.com>
 
 #include <simgear_config.h>
 #include <simgear/compiler.h>
+#include <thread>
 
 #ifdef HAVE_WINDOWS_H
 #include <windows.h>
@@ -42,11 +28,12 @@
 #include <signal.h>             // signal()
 #include <string.h>
 
-#include <iostream>
-#include <fstream>
-#include <string>
-#include <map>
 #include <algorithm>
+#include <fstream>
+#include <iostream>
+#include <map>
+#include <string>
+#include <chrono>
 
 #include <simgear/version.h>
 
@@ -66,10 +53,13 @@
 #include <simgear/io/HTTPRepository.hxx>
 #include <simgear/io/DNSClient.hxx>
 #include <simgear/structure/exception.hxx>
-#include <simgear/math/sg_random.h>
+#include <simgear/math/sg_random.hxx>
+#include <simgear/debug/ErrorReportingCallback.hxx>
 
 using namespace simgear;
-using namespace std;
+using std::string;
+using namespace std::string_literals;
+using namespace std::chrono_literals;
 
 namespace UpdateInterval
 {
@@ -79,7 +69,7 @@ namespace UpdateInterval
     static const double FailedAttempt     = 10*60;
 }
 
-typedef map<string,time_t> TileAgeCache;
+typedef std::map<std::string, time_t> TileAgeCache;
 
 ///////////////////////////////////////////////////////////////////////////////
 // helper functions ///////////////////////////////////////////////////////////
@@ -146,6 +136,37 @@ public:
     Status _status;
 };
 
+std::ostream& operator << (std::ostream& out, const SyncItem::Type& t)
+{
+    if (t == SyncItem::Stop)            return out << "Stop";
+    if (t == SyncItem::Tile)            return out << "Tile";
+    if (t == SyncItem::AirportData)     return out << "AirportData";
+    if (t == SyncItem::SharedModels)    return out << "SharedModels";
+    if (t == SyncItem::AIData)          return out << "AIData";
+    if (t == SyncItem::OSMTile)         return out << "OSMTile";
+    return out << ((int) t);
+}
+
+std::ostream& operator << (std::ostream& out, const SyncItem::Status& s)
+{
+    if (s == SyncItem::Invalid)     return out << "Invalid";
+    if (s == SyncItem::Waiting)     return out << "Waiting";
+    if (s == SyncItem::Cached)      return out << "Cached";
+    if (s == SyncItem::Updated)     return out << "Updated";
+    if (s == SyncItem::NotFound)    return out << "NotFound";
+    if (s == SyncItem::Failed)      return out << "Failed";
+    return out << ((int) s);
+}
+
+std::ostream& operator << (std::ostream& out, const SyncItem& s)
+{
+    return out << "SyncItem:{_dir="
+            << s._dir << " _type="
+            << s._type << " _status="
+            << s._status
+            << "}";
+}
+
 ///////////////////////////////////////////////////////////////////////////////
 
 /**
@@ -178,7 +199,7 @@ static const unsigned int NUM_SYNC_SLOTS = 4;
 
 /**
  * @brief translate a sync item type into one of the available slots.
- * This provides the scheduling / balancing / prioritising between slots.
+ * This provides the scheduling / balancing / prioritizing between slots.
  */
 static unsigned int syncSlotForType(SyncItem::Type ty)
 {
@@ -358,7 +379,7 @@ public:
 
     void drainWaitingTiles();
 
-    // commond helpers between both internal and external models
+    // common helpers between both internal and external models
 
     SyncItem::Status isPathCached(const SyncItem& next) const;
     void updated(SyncItem item, bool isNewDirectory);
@@ -520,14 +541,22 @@ std::string SGTerraSync::WorkerThread::dnsSelectServerForService(const std::stri
         dnsClient.update(0);
     }
 
-    if( naptrRequest->entries.empty() ) {
-        SG_LOG(SG_TERRASYNC, SG_ALERT, "Warning: no DNS entry found for '" << _dnsdn << "' '" << naptrRequest->qservice << "'" );
+    if (naptrRequest->hasError()) {
+        simgear::reportFailure(simgear::LoadFailure::NetworkError, simgear::ErrorCode::TerraSync,
+                        "DNS query failed for DNSDN:"s + _dnsdn + ":" + naptrRequest->errorMessage());
         return {};
     }
 
+    if( naptrRequest->entries.empty() ) {
+        simgear::reportFailure(simgear::LoadFailure::NetworkError, simgear::ErrorCode::TerraSync,
+                        "No DNS entries found for DNSDN:"s + _dnsdn + " with service '"s + naptrRequest->qservice + "'"s);
+        return {};
+    }
+
+
     // walk through responses, they are ordered by 1. order and 2. preference
     // For now, only take entries with lowest order
-    // TODO: try all available servers in the order given by preferenc and order
+    // TODO: try all available servers in the order given by preference and order
     int order = naptrRequest->entries[0]->order;
 
     // get all servers with this order and the same (for now only lowest preference)
@@ -557,8 +586,17 @@ std::string SGTerraSync::WorkerThread::dnsSelectServerForService(const std::stri
     // now pick a random entry from the available servers
     auto idx = static_cast<int>(sg_random() * availableServers.size());
     const auto server = availableServers.at(idx)->regexp;
-    SG_LOG(SG_TERRASYNC, SG_INFO, "picking entry # " << idx << ", server is " << server.substr(6, server.length() - 7););
-    return server.substr(6, server.length() - 7);
+    std::string ret = server.substr(6, server.length() - 7);
+    SG_LOG(SG_TERRASYNC, SG_INFO, "service=" << service << " returning entry # " << idx << ": " << ret);
+    {
+        std::string env = "SIMGEAR_TERRASYNC_SERVER_" + service;
+        const char* ret = getenv(env.c_str());
+        if (ret) {
+            SG_LOG(SG_TERRASYNC, SG_INFO, "service=" << service << " overriding to return " << env << " = " << ret);
+            return ret;
+        }
+    }
+    return ret;
 }
 
 void SGTerraSync::WorkerThread::run()
@@ -708,7 +746,7 @@ bool SGTerraSync::WorkerThread::beginSyncTile(SyncSlot& slot)
 {
     // avoid 404 requests by doing a sync which excludes all paths
     // except our tile path. In the case of a missing 1x1 tile, we will
-    // stop becuase all directories are filtered out, which is what we want
+    // stop because all directories are filtered out, which is what we want
 
     auto comps = strutils::split(slot.currentItem._dir, "/");
     if (comps.size() != 3) {
@@ -798,6 +836,7 @@ void SGTerraSync::WorkerThread::runInternal()
 
                 SG_LOG(SG_TERRASYNC, SG_INFO, "terrasync scenery provider of the day is '" << _httpServer << "'");
             } else {
+                std::this_thread::sleep_for(1s);
                 std::lock_guard<std::mutex> g(_stateLock);
                 _state._consecutive_errors++;
             }
@@ -858,7 +897,9 @@ SyncItem::Status SGTerraSync::WorkerThread::isPathCached(const SyncItem& next) c
         ii = _notFoundItems.find( next._dir );
         // Invalid means 'not cached', otherwise we want to return to
         // higher levels the cache status
-        return (ii == _notFoundItems.end()) ? SyncItem::Invalid : SyncItem::NotFound;
+        SyncItem::Status ret = (ii == _notFoundItems.end()) ? SyncItem::Invalid : SyncItem::NotFound;
+        SG_LOG(SG_TERRASYNC, SG_DEBUG, "next=" << next << " returning ret=" << ret);
+        return ret;
     }
 
     // check if the path still physically exists. This is needed to
@@ -866,11 +907,14 @@ SyncItem::Status SGTerraSync::WorkerThread::isPathCached(const SyncItem& next) c
     SGPath p(_local_dir);
     p.append(next._dir);
     if (!p.exists()) {
+        SG_LOG(SG_TERRASYNC, SG_DEBUG, "next=" << next << " returning SyncItem::Invalid");
         return SyncItem::Invalid;
     }
 
     time_t now = time(0);
-    return (ii->second > now) ? SyncItem::Cached : SyncItem::Invalid;
+    SyncItem::Status ret = (ii->second > now) ? SyncItem::Cached : SyncItem::Invalid;
+    SG_LOG(SG_TERRASYNC, SG_DEBUG, "next=" << next << " returning ret=" << ret);
+    return ret;
 }
 
 void SGTerraSync::WorkerThread::fail(SyncItem failedItem)
@@ -879,7 +923,7 @@ void SGTerraSync::WorkerThread::fail(SyncItem failedItem)
     time_t now = time(0);
 
     if (_osmCityServer.empty() && (failedItem._type == SyncItem::OSMTile)) {
-        // don't count these as errors, otherwise normla sync will keep
+        // don't count these as errors, otherwise normal sync will keep
         // being abandoned
     } else {
         _state._consecutive_errors++;
@@ -896,8 +940,8 @@ void SGTerraSync::WorkerThread::fail(SyncItem failedItem)
 
 void SGTerraSync::WorkerThread::notFound(SyncItem item)
 {
-    // treat not found as authorative, so use the same cache expiry
-    // as succesful download. Important for MP models and similar so
+    // treat not found as authoritative, so use the same cache expiry
+    // as successful download. Important for MP models and similar so
     // we don't spam the server with lookups for models that don't
     // exist
 
@@ -938,15 +982,20 @@ void SGTerraSync::WorkerThread::drainWaitingTiles()
     while (!waitingTiles.empty()) {
         SyncItem next = waitingTiles.pop_front();
         SyncItem::Status cacheStatus = isPathCached(next);
+        SG_LOG(SG_TERRASYNC, SG_INFO, "next._type=" << next._type
+                << " next._dir=" << next._dir
+                << " cacheStatus=" << cacheStatus
+                );
         if (cacheStatus != SyncItem::Invalid) {
             incrementCacheHits();
-            SG_LOG(SG_TERRASYNC, SG_BULK, "\nTerraSync Cache hit for: '" << next._dir << "'");
+            SG_LOG(SG_TERRASYNC, SG_BULK, "TerraSync Cache hit for: '" << next._dir << "'");
             next._status = cacheStatus;
             _freshTiles.push_back(next);
             continue;
         }
 
         const auto slot = syncSlotForType(next._type);
+        SG_LOG(SG_TERRASYNC, SG_INFO, "adding to _syncSlots slot=" << slot);
         _syncSlots[slot].queue.push_back(next);
     }
 }
@@ -1000,9 +1049,15 @@ void SGTerraSync::WorkerThread::initCompletedTilesPersistentCache() {
 
   for (int i = 0; i < cacheRoot->nChildren(); ++i) {
     SGPropertyNode *entry = cacheRoot->getChild(i);
-    bool isNotFound = (strcmp(entry->getName(), "not-found") == 0);
+    bool isNotFound = (entry->getNameString() == "not-found");
     string tileName = entry->getStringValue("path");
     time_t stamp = entry->getIntValue("stamp");
+    SG_LOG(SG_TERRASYNC, SG_DEBUG, "tileName=" << tileName
+            << " isNotFound=" << isNotFound
+            << " stamp=" << stamp
+            << " now=" << now
+            << " stamp<now=" << (stamp < now)
+            );
     if (stamp < now) {
       continue;
     }
@@ -1018,11 +1073,13 @@ void SGTerraSync::WorkerThread::initCompletedTilesPersistentCache() {
 void SGTerraSync::WorkerThread::writeCompletedTilesPersistentCache() const {
   // cache is disabled
   if (_persistentCachePath.isNull()) {
+    SG_LOG(SG_TERRASYNC, SG_DEBUG, "_persistentCachePath.isNull()");
     return;
   }
 
   sg_ofstream f(_persistentCachePath, std::ios::trunc);
   if (!f.is_open()) {
+    SG_LOG(SG_TERRASYNC, SG_DEBUG, "!f.is_open() _persistentCachePath=" << _persistentCachePath);
     return;
   }
 
@@ -1073,12 +1130,10 @@ void SGTerraSync::setRoot(SGPropertyNode_ptr root)
 {
     if (!root) {
         _terraRoot.clear();
-        _renderingRoot.clear();
         return;
     }
 
     _terraRoot = root->getNode("/sim/terrasync",true);
-    _renderingRoot = root->getNode("/sim/rendering", true);
 }
 
 void SGTerraSync::init()
@@ -1127,7 +1182,7 @@ void SGTerraSync::reinit()
             if (ns == "google") {
                 ns = "8.8.8.8";
             }
-            
+
             SG_LOG(SG_TERRASYNC,SG_INFO,"DNS server override:" << ns);
             _workerThread->setDNSServer(ns);
         }
@@ -1174,7 +1229,7 @@ void SGTerraSync::bind()
 
     //_terraRoot->getNode("use-built-in-svn", true)->setAttribute(SGPropertyNode::USERARCHIVE,false);
     //_terraRoot->getNode("use-svn", true)->setAttribute(SGPropertyNode::USERARCHIVE,false);
-    _terraRoot->getNode("intialized", true)->setBoolValue(true);
+    _terraRoot->getNode("initialized", true)->setBoolValue(true);
 
     // stalled is used as a signal handler (to connect listeners triggering GUI pop-ups)
     _stalledNode = _terraRoot->getNode("stalled", true);
@@ -1208,7 +1263,6 @@ void SGTerraSync::unbind()
     _stalledNode.clear();
     _activeNode.clear();
     _cacheHits.clear();
-    _renderingRoot.clear();
     _busyNode.clear();
     _updateCountNode.clear();
     _errorCountNode.clear();
@@ -1286,27 +1340,9 @@ void SGTerraSync::syncAirportsModels()
   _workerThread->request(a);
 }
 
-string_list SGTerraSync::getSceneryPathSuffixes() const
-{
-    string_list scenerySuffixes;
-
-    for (auto node : _renderingRoot->getChildren("scenery-path-suffix")) {
-        if (node->getBoolValue("enabled", true) && node->hasValue("name")) {
-            scenerySuffixes.push_back(node->getStringValue("name"));
-        }
-    }
-
-    if (scenerySuffixes.empty()) {
-        // if preferences didn't load, use some default
-        scenerySuffixes = {"Objects", "Terrain"}; // default values
-    }
-
-    return scenerySuffixes;
-}
-
 bool isOSMSuffix(const std::string& suffix)
 {
-    return (suffix == "Buildings") || (suffix == "Roads") || (suffix == "Pylons") || (suffix == "Details");
+    return (suffix == "Buildings") || (suffix == "Roads") || (suffix == "Pylons") || (suffix == "Details") || (suffix == "Trees");
 }
 
 void SGTerraSync::syncAreaByPath(const std::string& aPath)
@@ -1315,7 +1351,11 @@ void SGTerraSync::syncAreaByPath(const std::string& aPath)
         return;
     }
 
-    for (const auto& suffix : getSceneryPathSuffixes()) {
+    if (_sceneryPathSuffixes.empty()) {
+        throw sg_exception("SGTerraSync::syncAreaByPath called before scenery suffixes were defined.");
+    }
+
+    for (const auto& suffix : _sceneryPathSuffixes) {
         const auto dir = suffix + "/" + aPath;
         if (_workerThread->isDirActive(dir)) {
             continue;
@@ -1332,15 +1372,24 @@ bool SGTerraSync::scheduleTile(const SGBucket& bucket)
     return true;
 }
 
+void SGTerraSync::setSceneryPathSuffixes(const string_list& suffixes)
+{
+    _sceneryPathSuffixes = suffixes;
+}
+
 bool SGTerraSync::isTileDirPending(const std::string& sceneryDir) const
+{
+    return innerIsTileDirPending(sceneryDir, false);
+}
+
+bool SGTerraSync::innerIsTileDirPending(const std::string& sceneryDir, bool isOSM) const
 {
     if (!_workerThread->isRunning()) {
         return false;
     }
 
-    for (const auto& suffix : getSceneryPathSuffixes()) {
-        // don't wait on OSM dirs, even if enabled
-        if (isOSMSuffix(suffix)) {
+    for (const auto& suffix : _sceneryPathSuffixes) {
+        if (isOSM != isOSMSuffix(suffix)) {
             continue;
         }
 
@@ -1353,6 +1402,11 @@ bool SGTerraSync::isTileDirPending(const std::string& sceneryDir) const
     return false;
 }
 
+bool SGTerraSync::isTileDirPendingOSM(const std::string& sceneryDir) const
+{
+    return innerIsTileDirPending(sceneryDir, true);
+}
+
 void SGTerraSync::scheduleDataDir(const std::string& dataDir)
 {
     if (!_workerThread->isRunning()) {
@@ -1392,10 +1446,10 @@ void SGTerraSync::writeWarningFile(const SGPath& sceneryDir)
     os << "Any changes you make here or in sub-folders will be overwritten when TerraSync\n";
     os << "downloads updates.\n";
     os << "\n";
-    os << "To use custom scenery or data with FlightGear, put it in a differnet location\n";
+    os << "To use custom scenery or data with FlightGear, put it in a different location\n";
     os << "on your computer, then add the location using either the launcher 'Add-ons' page, or by\n";
     os << "passing '--fg-scenery=<location>' on the command line.";
-    os << endl;
+    os << std::endl;
 }
 
 
